# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip


import sys
import logging
import getpass
from collections import OrderedDict
import uuid

import numpy as np
import theano
import theano.tensor as T

from ..config import config
from .. import utils
from . import node_basic, neural, loss
from . import optimiser
from . import graphmanager
from . import graphutils

logger = logging.getLogger('elektronn2log')

user_name = getpass.getuser()

if sys.version_info.major >= 3:
    unicode = str

__all__ = ['Model', 'modelload', 'kernel_lists_from_node_descr',
           'params_from_model_file', 'rebuild_model', 'simple_cnn']
#TODO break all feed backs
#TODO docstrings for rebuild etc.

class Model(graphmanager.GraphManager):
    """
    Represents a neural network model and its current training state.

    The Model is defined and checked by running Model.designate_nodes() with
    appropriate Nodes as arguments (see example in examples/numa_mnist.py).

    Permanently saving a Model with its respective training state is possible
    with the Model.save() function.
    Loading a Model from a file is done by elektronn2.neuromancer.model.modelload().

    During training of a neural network, you can access the current Model
    via the interactive training shell as the variable "model"
    (see elektronn2.training.trainutils.user_input()).
    There are several statistics and hyperparameters of the Model that you
    can inspect and set directly in the shell, e.g. entering
    >>> model.lr = 1e-3
    and exiting the prompt again effectively sets the learning rate to 1e-3
    for further training.
    (This can also be done with the shortcut "setlr 1e-3".)
    """
    def __init__(self, name=""):
        super(Model, self).__init__(name=name)

        self.batch_size = None
        self.ndim = None
        self._desig_descr = dict()

        # Tracking / Monitoring etc
        self.iterations   = 0
        self.elapsed_time = 0
        self._last_exec_times = utils.CircularBuffer(
            config.time_per_step_smoothing_length)

        self._last_losses = utils.CircularBuffer(config.loss_smoothing_length)

        # Node stuff
        self.prediction_node = None
        self.prediction_ext = None
        self._prediction_ext_func = None

        self.loss_node = None
        self.target_node = None
        self.error_node = None
        self.input_node = None

        self.trainable_params = None
        self.nontrainable_params = None
        self._grad = None
        self._grad_inp = None
        self._grad_func = None
        self.optimisers = dict()
        self.debug_outputs = []

        self._activation_func = None

    def designate_nodes(self, input_node='input', target_node=None,
                        loss_node=None, prediction_node=None,
                        prediction_ext=None, error_node=None,
                         debug_outputs=None):
        """
        Register input, target and other special Nodes in the Model.

        Most of the Model's attributes are derived from the Nodes that are
        given as arguments here.
        """

        def designate_nodes(purpose, name): # Does nothing if name is None
            if isinstance(name, (list, tuple)):
                if purpose not in ['debug_outputs', 'prediction_ext']:
                    raise ValueError("Can only designate several nodes for "
                                     "debug outputs and prediction_ext")
                name = [n if isinstance(n, (str,unicode)) else n.name for n in name]
                nodes = []
                for n in name:
                    nodes.append(self.nodes[n])

                setattr(self, purpose, nodes)

            elif name:
                name = name if isinstance(name, (str, unicode)) else name.name
                setattr(self, purpose, self.nodes[name])

            self._desig_descr[purpose]=name # for reconstruction

        if debug_outputs is None:
            debug_outputs = []

        # Must use same names (second arg) as in kwargs of this function
        designate_nodes('input_node', input_node)
        designate_nodes('target_node', target_node)
        designate_nodes('loss_node', loss_node)
        designate_nodes('prediction_node', prediction_node)
        designate_nodes('error_node', error_node)

        designate_nodes('prediction_ext', prediction_ext)
        designate_nodes('debug_outputs', debug_outputs)

        # Prediction
        if self.prediction_node:
            self.batch_size = self.prediction_node.shape['b']
            self.ndim = self.prediction_node.shape.ndim
            if np.any(np.less(self.prediction_node.shape.fov, 0)): # If UpConvs contained
                in_sh = self.input_node.shape.spatial_shape
                sh = np.array(self.prediction_node.shape.spatial_shape)
                st = np.array(self.prediction_node.shape.strides)
                out_sh = st * (sh-1) + 1
                diff = np.subtract(in_sh, out_sh)
                if np.any(np.mod(diff, 2)):
                    raise ValueError("FOV is not centered. In_sh=%s, "
                    "out_sh*strides=%s, diff=%s"
                                     %(in_sh, out_sh, diff))
                self.prediction_node.shape._fov = diff # Hack
                self.target_node.shape._fov = diff     # Hack

            elif not self.prediction_node.shape.fov_all_centered:
                logger.warning("Not all field of views are centered (odd) "
                "this might cause problems for many setups")
            logger.info("Prediction properties:\n%s"\
                   %(self.prediction_node.shape.ext_repr))

            if not self.loss_node:
                n_param = self.prediction_node.all_params_count
                n_comp  = self.prediction_node.all_computational_cost
                n_comp_specific = float(n_comp)/self.prediction_node.shape.spatial_size
                logger.info("\nTotal Computational Cost of Model: {0:s}\n"
                            "Total number of trainable parameters: {1:,d}.\n"
                            "Computational Cost per pixel: {2:s}\n".
                            format(utils.pretty_string_ops(n_comp), n_param,
                                   utils.pretty_string_ops(n_comp_specific)))

        # Multiple output extended prediction
        if self.prediction_ext:
            outs = [n.output for n in self.prediction_ext]
            inp = graphutils.getinput_for_multioutput(self.prediction_ext)
            self._prediction_ext_func = graphutils.make_func(inp, outs,
                                                    name='Predictor Extended')
        # Loss, Model Parameters and their gradients, Optimisation
        if self.loss_node:
            self.trainable_params = list(self.loss_node.all_trainable_params.values()) # values of dict, not of shared!
            self.nontrainable_params = self.loss_node.all_nontrainable_params
            extra_updates          = self.loss_node.all_extra_updates

            self._grad = T.grad(self.loss_node.output, self.trainable_params,
                                disconnected_inputs="warn")
            self._grad_inp = self.loss_node.input_tensors
            self._grad_func = graphutils.make_func(self._grad_inp, self._grad,
                                                name='Gradient Func')

            debug_outputs = [n.output for n in self.debug_outputs]

            # Init *all* optimisers
            opt_init = (self._grad_inp, self.loss_node.output, self._grad,
                        self.trainable_params, extra_updates, debug_outputs)


            # TODO: self.optimisers should automatically register all subclasses of Optimiser
            self.optimisers = dict(SGD=optimiser.SGD(*opt_init),
                                   AdaGrad=optimiser.AdaGrad(*opt_init),
                                   AdaDelta=optimiser.AdaDelta(*opt_init),
                                   Adam=optimiser.Adam(*opt_init))

            n_param = self.loss_node.all_params_count
            n_comp  = self.loss_node.all_computational_cost
            if self.prediction_node:
                n_comp_specific = float(n_comp) / self.prediction_node.shape.spatial_size
            else:
                n_comp_specific = float(n_comp)/self.loss_node.shape.spatial_size
            logger.info("\nTotal Computational Cost of Model: {0:s}\n"
                        "Total number of trainable parameters: {1:,d}.\n"
                        "Computational Cost per pixel: {2:s}\n".
                        format(utils.pretty_string_ops(n_comp), n_param,
                               utils.pretty_string_ops(n_comp_specific)))


        # Activations (outputs os all intermediate layers
        activations = []
        activations_tt = []
        for node in self.nodes.values():
            # Split descriptors are stored as strings amongst the nodes
            if not isinstance(node, str) and not node.is_source:
                # multi output will apper in from-tensor nodes
                if not isinstance(node.output, (list,tuple)):
                    activations.append(node)
                    activations_tt.append(node.output)

        inp = graphutils.getinput_for_multioutput(activations)
        self._activation_func = graphutils.make_func(inp, activations_tt,
                                                        name='Activations')

    def save(self, file_name):
        """
        Save a Model (including its training state) to a pickle file.
        :param file_name: File name to save the Model in.
        """
        descriptors = self.serialise()
        utils.picklesave((descriptors, self._desig_descr), file_name)

    def loss(self, *args, **kwargs):
        return self.loss_node(*args, **kwargs)

    def gradients(self, *args, **kwargs):
        return self._grad_func(*args, **kwargs)

    def activations(self, *args, **kwargs):
       return self._activation_func(*args, **kwargs)

    def predict(self, *args, **kwargs):
        return self.prediction_node(*args, **kwargs)

    def predict_ext(self, *args, **kwargs):
        return self._prediction_ext_func(*args, **kwargs)

    def predict_dense(self, raw_img, as_uint8=False, pad_raw=False):
        return self.prediction_node.predict_dense(raw_img,
                                                  as_uint8=as_uint8,
                                                  pad_raw=pad_raw)

    def paramstats(self):
        print("Parameter statistics")
        for k, W in self.loss_node.all_trainable_params.items():
            W=W.get_value()
            print("Param %s:\tshape=%s,\tmean=%f,\tstd=%f,\tmedian(abs)=%f"\
                 %(k, W.shape, W.mean(), W.std(), np.median(np.abs(W))))


    def gradstats(self, *args, **kwargs):
        grads = self.gradients(*args, **kwargs)
        print("Gradient statistics")
        for g in grads:
            print("\tshape=%s,\tmean=%f,\tstd=%f,\tmedian(abs)=%f"\
                  %(g.shape, np.mean(g), np.std(g), np.median(np.abs(g))))

    def actstats(self, *args, **kwargs):
        acts = self.activations(*args, **kwargs)
        print("Activation statistics")
        for a in acts:
            print("\tshape=%s,\tmean=%f,\tstd=%f,\tmedian(abs)=%f"\
                  %(a.shape, np.mean(a), np.std(a), np.median(np.abs(a))))

    def set_opt_meta_params(self, opt_name, value_dict):
        self.optimisers[opt_name].set_opt_meta_params(value_dict)

    @property
    def lr(self):
        """
        Get learning rate.
        """
        return list(self.optimisers.values())[0].global_lr.get_value()

    @lr.setter
    def lr(self, val):
        """
        Set learning rate.
        """
        list(self.optimisers.values())[0].setlr(val)

    @property
    def mom(self):
        """
        Get momentum.
        """
        return list(self.optimisers.values())[0].global_mom.get_value()

    @mom.setter
    def mom(self, val):
        """
        Set momentum.
        """
        list(self.optimisers.values())[0].setmom(val)

    @property
    def wd(self):
        """
        Get weight decay.
        """
        return list(self.optimisers.values())[0].global_weight_decay.get_value()

    @wd.setter
    def wd(self, val):
        """
        Set weight decay.
        """
        list(self.optimisers.values())[0].setwd(val)

    @property
    def mixing(self):
        """
        Get mixing weights.
        """
        if isinstance(self.loss_node, loss.AggregateLoss):
            return self.loss_node.mixing_weights.get_value()
        else:
            try:
                for n in self.nodes.values():
                    if isinstance(n, loss.AggregateLoss):
                        break

                logger.info("model.loss_node is not of type AggregateLoss and "
                               "hence has no mixing_weights. But '%s' is, so this is "
                               "used." %n.name)
                return n.mixing_weights.get_value()
            except:
                logger.error("no mixing_weights found in model")

    @mixing.setter
    def mixing(self, val):
        """
        Set mixing weights.
        """
        val = graphutils.as_floatX(val)
        if isinstance(self.loss_node, loss.AggregateLoss):
            self.loss_node.mixing_weights.set_value(val)
        else:
            try:
                for n in self.nodes.values():
                    if isinstance(n, loss.AggregateLoss):
                        break

                logger.info("model.loss_node is not of type AggregateLoss and "
                               "hence has no mixing_weights. But '%s' is, so this is "
                               "used." % n.name)
                n.mixing_weights.set_value(val)
            except:
                logger.error("no mixing_weights found in model")

    @property
    def dropout_rates(self):
        """
        Get dropout rates.
        """
        dropout_rates = []
        for node in self.nodes.values():
            r = node.params.get('dropout_rate', None)
            if r:
                dropout_rates.append(r.get_value())

        return np.array(dropout_rates)

    @dropout_rates.setter
    def dropout_rates(self, rates):
        """
        Set dropout rates.

        If the argument is a number, all Nodes that support it are given this
        dropout rate.
        If the argument is a tuple, list or array, the Nodes in the Model
        that support it are given the rates in their respective ordering.
        """
        i = 0
        for node in self.nodes.values():
            r = node.params.get('dropout_rate', None)
            if r:
                if isinstance(rates, (tuple, list, np.ndarray)):
                    r.set_value(graphutils.as_floatX(rates[i]))
                else:
                    r.set_value(graphutils.as_floatX(rates))
                i += 1

    @property
    def gradnet_rates(self):
        """
        Get gradnet rates.

        Description: https://arxiv.org/abs/1511.06827
        """
        gradnet_rates = []
        for node in self.nodes.values():
            r = node.params.get('gradnet_rate', None)
            if r:
                gradnet_rates.append(r.get_value())

        return gradnet_rates

    @gradnet_rates.setter
    def gradnet_rates(self, rates):
        """
        Set gradnet rates.

        Description: https://arxiv.org/abs/1511.06827

        If the argument is a number, all Nodes that support it are given this
        gradnet rate.
        If the argument is a tuple, list or array, the Nodes in the Model
        that support it are given the rates in their respective ordering.
        """
        i = 0
        for node in self.nodes.values():
            r = node.params.get('gradnet_rate', None)
            if r:
                if isinstance(rates, (tuple, list, np.ndarray)):
                    r.set_value(graphutils.as_floatX(rates[i]))
                else:
                    r.set_value(graphutils.as_floatX(rates))
                i += 1

    @property
    def batch_normalisation_active(self):
        """
        Check if batch normalisation is active in any Node in the Model.
        """
        active = 0
        for node in self.nodes.values():
            if hasattr(node, 'batch_normalisation'):
                if node.batch_normalisation in ['train', 'fadeout']:
                    active += 1

        return (active > 0)


    @property
    def debug_output_names(self):
        """
        If debug_outputs is set, a list of all debug output names is returned.
        """
        if self.debug_outputs:
            return [x.name for x in self.debug_outputs]
        else:
            return None


    @property
    def prediction_feature_names(self):
        """
        If a prediction node is set, return its feature names.
        """
        ret = None
        if self.prediction_node:
            ret = self.prediction_node.feature_names

        return ret




    def get_param_values(self, skip_const=False, as_list=False):
        """
        Only use this to save/load parameters!

        Returns a dict of mapping the values of the params
        (such that they can be saved to disk)
        :param skip_const: whether to exclude constant parameters
        """
        p_dict = OrderedDict()
        for name,node in self.nodes.items():
            p_dict[name] = node.get_param_values(skip_const)

        ret = list(p_dict.values()) if as_list else p_dict
        return ret


    def set_param_values(self, value_dict, skip_const=False):
        """
        Only use this to save/load parameters!

        Sets new values for non constant parameters
        :param value_dict: dict mapping values by parameter name / or file name thereof
        :param skip_const: if dict also maps values for constants, these
        can be skipped, otherwise an exception is raised.
        """
        logger.info("Setting model parameters")
        if isinstance(value_dict, str):
            value_dict = utils.pickleload(value_dict)

        if isinstance(value_dict, dict):
            for k,v in value_dict.items():
                if k not in self.nodes:
                    raise KeyError("Graph Manager has no node %s" %(k,))
                self.nodes[k].set_param_values(v, skip_const)
        else:
            for p,n in zip(value_dict, self.nodes.values()):
                if not isinstance(n, str):
                    n.set_param_values(p, skip_const)

    @property
    def loss_input_shapes(self):
        """
        Get shape(s) of loss nodes' input node(s).

        The return value is either a shape (if one input) or a list of shapes
        (if multiple inputs).
        """
        sources = self.loss_node.input_nodes
        if len(sources)==0:
            return sources[0].shape
        else:
            return [s.shape for s in sources]

    @property
    def time_per_step(self):
        """
        Get average run time per training step.

        The average is calculated over the last n steps, where n is defined by
        the config variable time_per_step_smoothing_length (default: 50).
        """
        return self._last_exec_times.mean() + 1e-6

    @property
    def loss_smooth(self):
        """
        Get average loss during the last training steps.

        The average is calculated over the last n steps, where n is defined by
        the config variable time_per_step_smoothing_length (default: 50).
        """
        return self._last_losses.mean()


    def trainingstep(self, *args, **kwargs):
        """
        Perform one optimiser iteration.
        Optimizers can be chosen by the kwarg ``optimiser``.

        **Signature**: ``trainingstep(data, target(, *aux)(, **kwargs**))``

        Parameters
        ----------
        *args
            * data: floatX array
                  input [bs, ch (, z, y, x)]
            * targets: int16 array
                  [bs,((z,)y,x)] (optional)
            * (optional) other inputs: np.ndarray
                  depending in model

        **kwargs
            * optimiser: str
                  Name of the chosen optimiser class in
                  :py:mod:`elektronn2.neuromancer.optimiser`
            * update_loss: Bool
                 determine current loss *after* update step
                 (e.g. needed for queue, but ``get_loss`` can also be
                 called explicitly)

        Returns
        -------
        loss: floatX
            Loss (nll, squared error etc...)
        t: float
            Time spent on the GPU per step
        """

        opt_name = kwargs.get('optimiser', 'SGD')
        if opt_name not in self.optimisers:
            logger.warning("No optimiser '%s'. Falling back to SGD"%(opt_name,))

        ret = self.optimisers[opt_name](*args)
        loss = ret[0]
        if kwargs.get('update_loss', False):
            loss = self.loss(*args)

        t  = self.optimisers[opt_name].last_exec_time
        self.elapsed_time += t
        self._last_exec_times.append(t + 1e-10) # add some epsilon to ensure > 0
        self._last_losses.append(loss)
        self.iterations += 1

        if len(ret)>1:
            return loss, t, ret[1:]
        else:
            return loss, t, None

    def test_run_prediction(self):
        """
        Execute test run on the prediction node.
        """
        self.prediction_node.test_run()

    def measure_exectimes(self, n_samples=5, n_warmup=4, print_info=True):
        """
        Return an OrderedDict that maps node names to their estimated execution times in milliseconds.

        Parameters are the same as in elektronn2.neuromancer.node_basic.Node.measure_exectime()
        """
        exectimes = OrderedDict()
        for name, node in self.nodes.items():
            exectime = node.measure_exectime(n_samples=n_samples, n_warmup=n_warmup, print_info=print_info)
            # exec_time = node.local_exec_time
            exectimes[name] = exectime
        return exectimes

###############################################################################

def modelload(file_name, override_mfp_to_active=False,
              imposed_patch_size=None, imposed_batch_size=None,
              name=None, **model_load_kwargs):
    """
    Load a Model from a pickle file (created by Model.save()).

    model_load_kwargs: remove_bn, make_weights_constant (True/False)
    """  # TODO: Document params

    logger.info("Loading model from %s" %file_name)
    node_descr, desig_descr = utils.pickleload(file_name)

    if 'input_node' not in desig_descr:
        if override_mfp_to_active or imposed_patch_size is not None or\
            imposed_batch_size is not None:
            raise("To use 'override_mfp_to_active' or 'imposed_patch_size', the"
                  " saved model must have a designated 'input_node'")
    else:
        input_node_descr = node_descr[desig_descr['input_node']][0]
        if isinstance(input_node_descr.args[1], graphutils.TaggedShape):
            old_shape = input_node_descr.args[1] # Input node is "FromTensor"-Type
        else:
            old_shape = graphutils.TaggedShape(input_node_descr.args[0],
                                       input_node_descr.args[1],
                                       strides=input_node_descr.kwargs.get('strides', None),
                                       fov=input_node_descr.kwargs.get('fov', None))
        old_patch_size = old_shape.spatial_shape

        ### Changing the shape ###
        changed = False
        if imposed_batch_size is not None:
            changed = True
            old_shape = old_shape.updateshape('b', imposed_batch_size)

        filter_shapes, pool_shapes, mfps = kernel_lists_from_node_descr(node_descr)
        if override_mfp_to_active:

            if imposed_patch_size is None:
                imposed_patch_size = old_patch_size
                # raise ValueError("'override_mfp_to_active' is True but the "
                #                  "patch_size is not updated. This does not work.")

            mfps = [True for mfp in mfps]
            model_load_kwargs['override_mfp_to_active'] = True

            ### Inject a FragmentsToDense node prior to softmax ###
            from elektronn2.neuromancer.neural import FragmentsToDense
            pred_name = desig_descr['prediction_node']
            pred_descr = node_descr[pred_name][0]
            ###TODO inject after prediction_node because this does not work for Concat nodes as prediction node. But then we need to change desig_descr???
            # create descriptor for a FragmentsToDense prior to softmax node
            # make the 'old' parent of pred, the parent of the reshape
            guard = str(uuid.uuid4())[:5]
            reshape_descr = graphmanager.NodeDescriptor([pred_descr.args[0]],
                                           dict(name='to_dense'+guard),
                                           FragmentsToDense, None)
            reshape_descr = [reshape_descr, OrderedDict()]
            pred_descr.args[0] = graphmanager.NodePointer('to_dense'+guard) # mutate parent

            new_model_descr = OrderedDict()
            for k,v in node_descr.items():
                if k == pred_name:
                    new_model_descr['to_dense'+guard] = reshape_descr

                new_model_descr[k] = v

            node_descr = new_model_descr

        if imposed_patch_size is not None:
            if len(imposed_patch_size)!=len(old_patch_size):
                raise ValueError("The dimensionality of the model and the "
                                 "imposed patchsize do not match.")

            if not np.array_equal(imposed_patch_size, old_patch_size) or override_mfp_to_active:
                changed = True
                ndim = len(filter_shapes[0])
                if not filter(lambda x: x[0].cls==neural.UpConv,
                              filter(lambda x: not isinstance(x, graphmanager.SplitDescriptor),
                                     node_descr.values())):

                    valid_patch_size = utils.get_cloesest_valid_patch_size(filter_shapes,
                                                     pool_shapes,
                                                     desired_patch_size=imposed_patch_size,
                                                     mfp=mfps,
                                                     ndim=ndim)

                else:
                    valid_patch_size = imposed_patch_size
                    logger.warning("Impose patch size is not failsave for UpConvs")
                for i, sh in enumerate(valid_patch_size):
                    old_shape = old_shape.updateshape(old_shape.spatial_axes[i], sh)

        if changed:
            input_node_descr.args[0] = tuple(old_shape.shape)
            logger.warning("Warning: the input shape of the image input node "
            "was changed during modelload, this change is NOT reflected in "
            "the printed shapes of other nodes (e.g. labels), but they "
            "have changed accordingly! This should be fixed in future, "
            "but the model can now be used for predictions nonetheless.")


    model = node_basic.model_manager.newmodel(name)
    model.restore(node_descr, **model_load_kwargs)
    if desig_descr:
        model.designate_nodes(**desig_descr)

    return model


def inject_source(node_descr, child_name, shape, tags, strides=None,
                  fov=None, dtype=theano.config.floatX, name='input', parent_no=0):
    logger.warning("'inject_source' is unsave: it works only for nodes that have"
                   "1 child and this child must have 1 parent. You are trying"
                   "to insert a source node prior to %s" %child_name)
    from elektronn2.neuromancer.node_basic import Input
    #name = name + '_' + str(uuid.uuid4())[:5] maybe not needed...?
    input_descr = graphmanager.NodeDescriptor((shape, tags),
                                dict(strides=strides, fov=fov, dtype=dtype,
                                     name=name),
                                Input, None)
    input_descr = [input_descr, OrderedDict()]

    child_descr = node_descr[child_name][0]
    # mutate parent of child
    if hasattr(child_descr.args[0], '__len__'):
        p_name = child_descr.args[0][parent_no]
        child_descr.args[0][parent_no] = graphmanager.NodePointer(name)
    else:
        p_name = child_descr.args[parent_no]
        child_descr.args[parent_no] = graphmanager.NodePointer(name)

    logger.warning("'inject_source' replacing %s input of node %s" %(p_name, child_name))
    new_model_descr = OrderedDict()
    for k, v in node_descr.items():
        if k==child_name:
            new_model_descr[name] = input_descr

        new_model_descr[k] = v

    return new_model_descr, name

def inject_source_split(node_descr, split_name, shape, tags, strides=None,
                      fov=None, dtype=theano.config.floatX, name='input'):

    logger.warning("'inject_source_split' is unsave")
    from elektronn2.neuromancer.node_basic import Input

    input_descr = graphmanager.NodeDescriptor((shape, tags),
                                              dict(strides=strides, fov=fov,
                                                   dtype=dtype,
                                                   name=name),
                                              Input, None)
    input_descr = [input_descr, OrderedDict()]

    split_descr = node_descr[split_name]
    split_descr.node_id = name

    new_model_descr = OrderedDict()
    for k, v in node_descr.items():
        if k==split_name:
            new_model_descr[name] = input_descr

        new_model_descr[k] = v

    return new_model_descr, name

def rebuild_rnn(model):
    """
    Rebuild an RNN by saving it to a file and reloading it from there.

    :param model: Model object.
    :return: Rebuilt Model.
    """
    img_shape = model.input_node.shape.shape
    img_tags  = model.input_node.shape.tags
    trace_it = model['scan'].n_steps
    descriptors = model.serialise()

    if 'conv_r' in descriptors:
        descriptors, input_name = inject_source_split(descriptors, 'split', img_shape,img_tags, name='img')
    else:
        descriptors, input_name = inject_source(descriptors, 'conv', img_shape, img_tags, name='img')


    if 'trace_join' in model.nodes:
        descriptors, _ = inject_source(descriptors, 'trace_join', (1,3*trace_it), 'b,f', name='trace_feedback')
    if isinstance(model['mem_hid'], node_basic.ValueNode):
        sh = model['mem_hid'].shape.shape
        descriptors, _ = inject_source(descriptors, 'gru', sh, 'b,f', name='mem_hid_feedback', parent_no=1)

    desig_descr = dict(model._desig_descr) # copy because we destroy
    desig_descr['input_node'] = input_name
    utils.picklesave((descriptors, desig_descr),
                     '/tmp/%s-tempmodel.pkl'%user_name)
    new_model = modelload('/tmp/%s-tempmodel.pkl' %user_name)
    return new_model


def rebuild_decoder(model, state_name, state_child_name):
    shape = model[state_name].shape.shape
    tags  = model[state_name].shape.tags
    descriptors = model.serialise()
    descriptors, input_name = inject_source(descriptors, state_child_name,
                                            shape, tags, name='state_input')

    desig_descr = dict(model._desig_descr) # copy because we destroy
    desig_descr['input_node'] = input_name

    # new_model_descr = OrderedDict()
    # skip = True
    # for k, v in descriptors.items():
    #     if k==input_name:
    #         skip = False
    #
    #     if not skip:
    #         new_model_descr[k] = v
    # Works only if a stuff from desig_descr is removed

    utils.picklesave((descriptors, desig_descr),
                     '/tmp/%s-tempmodel.pkl'%user_name)
    new_model = modelload('/tmp/%s-tempmodel.pkl' %user_name)
    return new_model



def rebuild_model(model, override_mfp_to_active=False,
                 imposed_patch_size=None, name=None, **model_load_kwargs):
    """
    Rebuild a Model by saving it to a file and reloading it from there.

    :param model: Model object.
    :param override_mfp_to_active: (See elektronn2.neuromancer.model.modelload()).
    :param imposed_patch_size: (See elektronn2.neuromancer.model.modelload()).
    :param name: New model name.
    :param model_load_kwargs:  Additional kwargs for restoring Model
        (see elektronn2.neuromancer.graphmanager.GraphManager.restore()).
    :return: Rebuilt Model.
    """
    model.save('/tmp/%s-tempmodel.pkl'%user_name)
    new_model = modelload('/tmp/%s-tempmodel.pkl' %user_name,
                          override_mfp_to_active=override_mfp_to_active,
                          imposed_patch_size=imposed_patch_size,
                          name=name, **model_load_kwargs)

    return new_model


def kernel_lists_from_node_descr(model_descr):
    """
    Extract the tuple (filter_shapes, pool_shapes, mfp) from a model description.

    :param model_descr: Model description OrderedDict.
    :return: Tuple (filter_shapes, pool_shapes, mfp).
    """
    filter_shapes = []
    pool_shapes   = []
    mfp           = []
    logger.warning("Warning [kernel_lists_from_model_descr]:"
                   "kernel properties are only considered for 'Conv' "
                   "nodes (excluding 'UpConv'!)")
    for name, descr in model_descr.items():
        if isinstance(descr, (list, tuple)):
          if descr[0].cls.__name__ == 'Conv':
              filter_shapes.append(descr[0].args[2])
              pool_shapes.append(descr[0].args[3])
              mfp.append(descr[0].kwargs.get('mfp', False))

    return filter_shapes, pool_shapes, mfp

def params_from_model_file(file_name):
    """
    Load parameters from a model file.

    :param file_name: File name of the pickled Model.
    :return: OrderedDict of model parameters.
    """
    logger.info("Extracting parameters from %s" %file_name)
    node_descr, desig_descr = utils.pickleload(file_name)
    params = OrderedDict()
    for name, descr in node_descr.items():
        if isinstance(descr, (list, tuple)) and len(descr[1]):
          params[name] = descr[1]

    return params


def simple_cnn(batch_size, n_ch, n_lab, desired_input, filters, nof_filters,
               activation_func, pools, mfp=False, tags=None,name=None):
    """
    Create a simple Model of a convolutional neural network.
    :param batch_size: Batch size (how many data samples are used in one
        update step).
    :param n_ch: Number of channels.
    :param n_lab: Number of distinct labels (classes).
    :param desired_input: Desired input image size. (Must be smaller than the
        size of the training images).
    :param filters: List of filter sizes in each layer.
    :param nof_filters: List of number of filters for each layer.
    :param activation_func: Activation function.
    :param pools: List of maxpooling factors for each layer.
    :param mfp: List of bools that tell if max fragment pooling should be used
        in each layer (only intended for prediction).
    :param tags: Tuple of tags for Input node (see docs of
        elektronn2.neuromancer.node_basic.Input).
    :param name: Name of the model.
    :return: Network Model.
    """  # TODO: params in docstring.

    ndim = len(desired_input)
    dimcalc = utils.cnncalculator(filters, pools, desired_input,
                         mfp=mfp, force_center=True, ndim=ndim)

    patch_size = dimcalc.input
    input_size = (batch_size, n_ch) + tuple(patch_size)

    node_basic.model_manager.newmodel(name)
    inp = node_basic.Input(input_size, tags)
    conv = list(zip(nof_filters, filters, pools, activation_func))
    for i,(n, f, p, act) in enumerate(conv):
        inp = neural.Conv(inp, n, f, p, mfp=mfp,
                         activation_func=act, )

    # last Layer
    out = neural.Conv(inp, config.n_lab, (1,)*ndim, (1,)*ndim,
                     activation_func='lin')
    if mfp:
        out = neural.FragmentsToDense(out)

    probs = loss.Softmax(out, n_class=n_lab, name='class_probabilities')

    # Label layer
    l_sh = out.shape.copy()
    l_sh.updateshape('f', 1)
    l     = node_basic.Input_like(l_sh, dtype='int16', name='labels')

    # Loss
    loss_pix  = loss.MultinoulliNLL(out, l, target_is_sparse=True)
    loss_scalar = loss.AggregateLoss(loss_pix, name='nll')

    # Model
    model = node_basic.model_manager.getmodel()
    model.designate_nodes(input_node=inp, target_node=l, loss_node=loss_scalar,
                                  prediction_node=probs,
                                  prediction_ext=[loss_scalar, loss_scalar, probs])
    return model



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved
from __future__ import absolute_import

from . import model

from .node_basic import *
from . import node_basic
from .neural import *
from . import neural
from .various import *
from . import various
from .loss import *
from . import loss
from . import variables
from .variables import *



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip


import logging
import time

import numpy as np
import theano
from theano import tensor as T, gof
from theano.gradient import disconnected_type

from .. import config
from .. import utils
from .node_basic import Node, GenericInput, FromTensor, model_manager
from .graphutils import TaggedShape, floatX
from .variables import VariableParam

logger = logging.getLogger('elektronn2log')
inspection_logger = logging.getLogger('elektronn2log-inspection')

__all__ = ['GaussianRV', 'SkelLoss',
           'SkelPrior', 'Scan', 'SkelGetBatch', 'SkelLossRec',
           'Reshape', 'SkelGridUpdate']


class GaussianRV(Node):
    """
    Parameters
    ----------
    mu: node
        Mean of the Gaussian density
    sig: node
        Sigma of the Gaussian density
    n_samples: int
        Number of samples to be drawn per instance.
        Special case '0': draw 1 sample but don't' increase rank of tensor!

    The output is a **sample** from separable Gaussians of given mean and
    sigma (but this operation is still differentiable, due to the
    "re-parameterisation trick").

    The output dimension mu.ndim+1 because the samples are accumulated along
    a new axis **right** of 'b' (batch).
    """

    def __init__(self, mu, log_sig, n_samples=0, name="state",
                 print_repr=True):
        super(GaussianRV, self).__init__((mu, log_sig), name, print_repr)

        self.mu = mu
        self.log_sig = log_sig
        self.n_samples = n_samples


    def _make_output(self):
        """ Computation of Theano Output """
        # It is assumed that all other dimensions are matching
        mu  = self.mu.output
        log_sig = self.log_sig.output
        sig = T.exp(log_sig)
        rng = T.shared_randomstreams.RandomStreams(int(time.time()))

        if self.n_samples>0:
            samples = []
            pattern = list(range(mu.type.ndim))
            batch_index = self.parent[0].shape.tag2index('b')
            pattern.insert(batch_index+1, 'x') # +1 because it is right of batch
            for i in range(self.n_samples):
                noise = rng.normal(mu.shape)
                z = mu + sig * noise
                samples.append(z.dimshuffle(pattern))
                z = T.concatenate(samples, axis=batch_index)
        else:
            noise = rng.normal(mu.shape)
            z = mu + sig * noise

        self.output = z

    def _calc_shape(self):
        if self.n_samples>0:
            self.shape = self.parent[0].shape.addaxis('b', self.n_samples, 's')
        else:
            self.shape = self.parent[0].shape.copy()

    def _calc_comp_cost(self):
        n = self.parent[0].shape.stripnone_prod
        if self.n_samples>0:
            self.computational_cost = n * self.n_samples
        else:
            self.computational_cost = n

    def make_priorlayer(self):
        """
        Creates a new Layer that calculates the Auto-Encoding-Variation-Bayes
        (AEVB) prior corresponding to this Layer.
        """
        # In particular the number of samples is the same, otherwise the relative
        # weight of prior and objective are wrong!
        return GaussianAEVBPrior(self.mu, self.log_sig,
                                 self.n_samples, name=self.name+"_prior")


###############################################################################

class GaussianAEVBPrior(GaussianRV):
    """
    Parameters
    ----------
    mu: Layer
        Mean of the Gaussian density
    sig: Layer
        Sigma of the Gaussian density
    n_samples: int
        Number of samples to be drawn per instance.
        Special case '0': draw 1 sample but don't' increase rank of tensor!

    The prior basically puts a L2-norm on mu and a similar constraint on sig
    but transformed such that sig=1 is favoured and for sig --> 0 it goes to inf.

    This Layer should only be created using the dedicated method of ``GaussianRV``.
    """

    def __init__(self, mu, log_sig, n_samples=0, name="prior",
                 print_repr=True):
        super(GaussianAEVBPrior, self).__init__(mu, log_sig, n_samples,
                                                name, print_repr)

    def _make_output(self):
        """ Computation of Theano Output """
        # It is assumed that all other dimensions are matching
        mu  = self.mu.output
        log_sig = self.log_sig.output
        sig = T.exp(log_sig)
        prior = 0.5 * (mu**2 + sig**2 - 1 - 2*log_sig)

        if self.n_samples>0:
            # The prior is replicated along a new sample axis in order to ensure
            # correct relative normalisation
            pattern = list(range(mu.type.ndim))
            batch_index = self.parent[0].shape.tag2index('b')
            pattern.insert(batch_index+1, 'x') # +1 because it is right of batch
            prior = prior.dimshuffle(pattern)
            prior = T.repeat(prior, self.n_samples, axis=-1)

        self.output = prior

###############################################################################

class StereographicMap(Node):
    """
    Interpret pred (bs, 3) as (X,Y,R), map to
    $(x, y, z) = R \cdot \left(\frac{2 X}{1 + X^2 + Y^2}, \frac{2 Y}{1 + X^2 + Y^2}, \frac{-1 + X^2 + Y^2}{1 + X^2 + Y^2}\right)$

    Parameters
    ----------
    pred
    name
    print_repr
    """

    def __init__(self, pred, name="stereo_map", print_repr=True):
        super(StereographicMap, self).__init__(pred, name, print_repr)
        self.pred   = pred.output
        self.pred_shape = pred.shape

    def _make_output(self):
        assert len(self.pred_shape)==2
        assert self.pred_shape[-1] == 3
        assert self.pred_shape[0] == 1

        R, Y, X = self.pred[:,0], self.pred[:,1], self.pred[:,2]
        R = T.nnet.softplus(R)

        z = - R * (-1 + X ** 2 + Y ** 2) / (1 + X ** 2 + Y ** 2)
        y = R * 2 * Y / (1 + X ** 2 + Y ** 2)
        x = R * 2 * X / (1 + X ** 2 + Y ** 2)

        self.output = T.stack((z, y, x)).T

    def _calc_shape(self):
        self.shape = TaggedShape([1,3], ['b','f'])


class SkelPrior(Node):
    """
    pred must be a vector of shape [(1,b),(3,f)] or [(3,f)]
    i.e. only batch_size=1 is supported.

    Parameters
    ----------
    pred
    target_length
    prior_n
    prior_posz
    prior_z
    prior_xy
    name
    print_repr
    """

    def __init__(self, pred, target_length=5.0, prior_n=0.0, prior_posz=0.0, prior_z=0.0, prior_xy=0.0,
                 name="skel_prior", print_repr=True):
        super(SkelPrior, self).__init__(pred, name, print_repr)

        self.pred = pred.output
        self.pred_shape = pred.shape
        self.target_length = VariableParam(value=target_length, name="target_length",
                                          dtype=floatX, apply_train=False)
        self.prior_n = VariableParam(value=prior_n, name="prior_n",
                                          dtype=floatX, apply_train=False)
        self.prior_z = VariableParam(value=prior_z, name="prior_z",
                                     dtype=floatX, apply_train=False)
        self.prior_posz = VariableParam(value=prior_posz, name="prior_posz",
                                     dtype=floatX, apply_train=False)
        self.prior_xy = VariableParam(value=prior_xy, name="prior_xy",
                                     dtype=floatX, apply_train=False)

        self.params['target_length'] = self.target_length
        self.params['prior_n'] = self.prior_n
        self.params['prior_z'] = self.prior_z
        self.params['prior_posz'] = self.prior_posz
        self.params['prior_xy'] = self.prior_xy


    def _make_output(self):
        assert len(self.pred_shape) in [1,2]
        assert self.pred_shape[-1] == 3
        if len(self.pred_shape)==2:
            assert self.pred_shape[0] == 1
            self.pred = self.pred[0]

        #self.pred = self.pred[:3] # just work on first 3 entries, rest might be anything

        aniso      = np.array([2,1,1], dtype=np.float32)
        norm       = T.sqrt(T.sum( (self.pred * aniso)**2))
        prior_n    = self.prior_n    * T.maximum(abs(norm - self.target_length) - 0.75, 0) # norm, maring of +/- 0.75
        prior_posz = self.prior_posz * T.nnet.softplus(-self.pred[0]) # penalise negative z
        prior_z    = self.prior_z    * abs(self.target_length/2-self.pred[0])  # make
        prior_xy   = self.prior_xy   *  T.sqrt(T.sum(self.pred[1:] ** 2)) # penalise larger xy

        self.output = prior_n + prior_posz + prior_z + prior_xy
        self._debug_outputs.extend([norm, prior_n, prior_posz,  prior_z, prior_xy])

    def _calc_shape(self):
        self.shape = TaggedShape([1], ['f'])


class SkelLossOP(theano.Op):
    """
    Parameters
    ----------
    predicted vec

    skel_obj

    Returns
    -------
    scalar loss
    """
    __props__ = ()
    def make_node(self, *inputs, **kwargs):
        inputs = list(inputs)
        inputs[0] = T.as_tensor_variable(inputs[0])
        loss = T.fvector('skel_loss')
        grad = T.fvector('skel_loss_grad')
        outputs = [loss, grad]
        self.loss_kwargs = kwargs
        return gof.Apply(self, inputs, outputs)

    def perform(self, node, inputs, output_storage):
        new_position_c, skel_obj, trafo = inputs
        loss, grad = output_storage

        new_position_l, _ = trafo.cnn_pred2lab_position(new_position_c)
        new_position_s = new_position_l[::-1]
        loss_, grad_s = skel_obj.get_loss_and_gradient(new_position_s,
                                         **self.loss_kwargs) # loggs nearest_s
        grad_l = grad_s[::-1]
        grad_c = trafo.lab_coord2cnn_coord(grad_l)

        loss[0], grad[0] = loss_, grad_c
        if config.inspection:
            inspection_logger.info("pred_s: %s, grad_s: %s, pred_c: %s, grad_c: %s"
                                   % (new_position_s.tolist(),
                                      grad_s.tolist(),
                                      new_position_c.tolist(),
                                      grad_c.tolist()))


    def grad(self, inputs, outputs_gradients):
        grad_0 = self(*inputs)[1]
        grad_1 = disconnected_type()
        grad_2 = disconnected_type()
        return [grad_0, grad_1, grad_2]

    def connection_pattern(self, node):
        # there is only a grad of the first output w.r.t to the first input
        return [[True, False],[False, False],[False, False]]


class SkelLossN(Node):
    """
    pred must be a vector of shape [(1,b),(3,f)] or [(3,f)]
    i.e. only batch_size=1 is supported

    Parameters
    ----------
    pred
    skel
    trafo
    loss_kwargs
    name
    print_repr
    """

    def __init__(self, pred, skel, trafo, loss_kwargs, name="skel_loss", print_repr=True):
        super(SkelLossN, self).__init__((pred, skel, trafo), name, print_repr)

        self.skel = skel.output
        self.trafo = trafo.output
        self.pred = pred.output
        self.pred_shape = pred.shape
        self.loss_kwargs = loss_kwargs

    def _make_output(self):
        assert len(self.pred_shape) in [1,2]
        assert self.pred_shape[-1] == 3
        if len(self.pred_shape)==2:
            assert self.pred_shape[0] == 1
            self.pred = self.pred[0]

        loss, grad = SkelLossOP()(self.pred, self.skel,
                                  self.trafo, **self.loss_kwargs)
        self.output = loss

    def _calc_shape(self):
        self.shape = TaggedShape([1], ['f'])


def SkelLoss(pred, loss_kwargs, skel=None, name="skel_loss", print_repr=True):
    if skel is None:
        skel = GenericInput(name='skeleton')  # This is not an argument to this node

    trafo = GenericInput(name='trafo')
    return SkelLossN(pred, skel, trafo, loss_kwargs, name=name, print_repr=print_repr)

###############################################################################

class SkelLossRecOP(theano.Op):
    """
    For Recurrent
    """
    __props__ = ()
    def make_node(self, *inputs, **kwargs):
        inputs = list(inputs)
        inputs[0] = T.as_tensor_variable(inputs[0])
        loss = T.fvector('skel_loss')
        grad = T.fvector('skel_loss_grad')
        outputs = [loss, grad]
        self.loss_kwargs = kwargs
        return gof.Apply(self, inputs, outputs)

    def perform(self, node, inputs, output_storage):
        new_position_c, pred_features, skel_obj = inputs
        new_position_l, tracing_direc_il = skel_obj.trafo.cnn_pred2lab_position(new_position_c)
        new_position_s = new_position_l[::-1]
        tracing_direc_is = tracing_direc_il[::-1]
        if config.inspection:
            inspection_logger.info("LossOP, node pred %s" %(np.array_str(pred_features, precision=2, suppress_small=True),))
            inspection_logger.info("LossOP, new_position_c: %s, new_position_l: %s"%(new_position_c,np.array_str(new_position_l, precision=1, suppress_small=True)))
        loss_, grad_s, nearest_s = skel_obj.step_feedback(new_position_s,
                                                          tracing_direc_is,
                                                          new_position_c,
                                                          pred_features,
                                                          **self.loss_kwargs)
        grad_l = grad_s[::-1]
        grad_c = skel_obj.trafo.lab_coord2cnn_coord(grad_l)
        loss, grad = output_storage
        loss[0] = loss_.astype(np.float32)
        grad[0] = grad_c.astype(np.float32)
        if config.inspection:
            inspection_logger.info("LossOP, loss: %s, grad: %s"%(loss, grad_c))


    def grad(self, inputs, outputs_gradients):
        grad_0 = self(*inputs)[1]
        grad_1 = disconnected_type()
        grad_2 = disconnected_type()
        return [grad_0, grad_1, grad_2]

    def connection_pattern(self, node):
        # there is only a grad of the first output w.r.t to the first input
        return [[True, False],[False, False],[False, False]]

class SkelLossRec(Node):
    """
    pred must be a vector of shape [(1,b),(3,f)] or [(3,f)]
    i.e. only batch_size=1 is supported.

    Parameters
    ----------
    pred
    skel
    loss_kwargs
    name
    print_repr
    """

    def __init__(self, pred, skel, loss_kwargs, name="skel_loss", print_repr=True):
        super(SkelLossRec, self).__init__((pred, skel), name, print_repr)

        self.skel = skel.output
        self.pred = pred.output
        self.pred_shape = pred.shape
        self.loss_kwargs = loss_kwargs

    def _make_output(self):
        assert len(self.pred_shape) in [1,2]
        assert self.pred_shape[-1] == 10
        if len(self.pred_shape)==2:
            assert self.pred_shape[0] == 1
            self.pred = self.pred[0]

        pred = self.pred[:3]
        pred_feat = self.pred[3:]

        loss, grad = SkelLossRecOP()(pred, pred_feat,
                                     self.skel, **self.loss_kwargs)
        self.output = loss

    def _calc_shape(self):
        self.shape = TaggedShape([1], ['f'])

###############################################################################

class SkelGridUpdateOP(theano.Op):
    """
    For Recurrent
    """
    __props__ = ()

    def make_node(self, *inputs, **kwargs):
        inputs = list(inputs)
        best_pred = T.fmatrix('best_pred')
        preds     = T.fmatrix('preds')
        scores    = T.fvector('scores')
        outputs   = [best_pred, preds, scores]
        self.kwargs = kwargs
        return gof.Apply(self, inputs, outputs)

    def perform(self, node, inputs, output_storage):
        grid, skel_obj, radius, bio = inputs
        best_pred_, preds_, scores_ = skel_obj.step_grid_update(grid, radius, bio)
        best_pred, preds, scores = output_storage
        best_pred[0] = best_pred_
        preds[0] = preds_
        scores[0] = scores_


    def grad(self, inputs, outputs_gradients):
        grad_0 = disconnected_type()
        grad_1 = disconnected_type()
        return [grad_0, grad_1]

    def connection_pattern(self, node):
        # there is only a grad of the first output w.r.t to the first input
        return [[True, False, False], [False, False, False]]


class SkelGridUpdateN(Node):
    """
    pred must be a vector of shape [(1,b),(3,f)] or [(3,f)]
    i.e. only batch_size=1 is supported

    Parameters
    ----------
    grid
    skel
    radius
    bio
    name
    print_repr
    """

    def __init__(self, grid, skel, radius, bio, name="grid2pred",
                 print_repr=True):
        super(SkelGridUpdateN, self).__init__((grid, skel, radius, bio), name, print_repr)

        self.skel = skel.output
        self.grid = grid.output
        self.radius = radius.output
        self.bio = bio.output

    def _make_output(self):
        best_pred, preds, scores = SkelGridUpdateOP()(self.grid, self.skel,
                                                      self.radius, self.bio)
        self.output = [best_pred, preds, scores]
        self.output_names = ['tracing', 'tracings', 'scores']

    def _calc_shape(self):
        pred_sh  = TaggedShape((1,3), 'b,f')
        preds_sh = TaggedShape((1, None, 3), 'b,s,f')
        scores   = TaggedShape((1, None), 'b,s')
        self.shape = [pred_sh, preds_sh, scores]

def skelgridupdate_split(parent, name='skelgridupdate_split'):
    args = ()
    kwargs = dict(name=name)
    model_manager.current.register_split(parent, skelgridupdate_split, name, args,
                                         kwargs)

    outs = []
    for out, out_sh, out_name in zip(parent.output, parent.shape,
                                     parent.output_names):
        outs.append(FromTensor(out, out_sh, parent,
                               name=out_name, print_repr=False))

    return outs


def SkelGridUpdate(grid, skel, radius, bio, name="skelgridupdate", print_repr=True):
    preds = SkelGridUpdateN(grid, skel, radius, bio, name=name, print_repr=print_repr)
    outputs = scansplit(preds, name='skelgridupdate_split')
    return outputs

###############################################################################



class SkelGetBatchOP(theano.Op):
    """
    Parameters
    ----------
    predicted vec

    skel_obj

    Returns
    -------
    scalar loss
    """
    __props__ = ()
    def make_node(self, *inputs, **kwargs):
        inputs = list(inputs)
        img         = T.TensorType(theano.config.floatX, (False,) * 5, name='image')()
        target_img  = T.TensorType(theano.config.floatX, (False,) * 5, name='target_img')()
        target_grid = T.TensorType(theano.config.floatX, (False,) * 5, name='target_grid')()
        target_node = T.TensorType(theano.config.floatX, (False,) * 2, name='target_node')()
        outputs = [img, target_img, target_grid, target_node]

        self.get_batch_kwargs = kwargs
        return gof.Apply(self, inputs, outputs)

    def perform(self, node, inputs, output_storage):
        skel_obj, prediction, scale_strenght = inputs
        img, target_img, target_grid, target_node = output_storage
        batch = skel_obj.getbatch(prediction, scale_strenght, **self.get_batch_kwargs)
        img_, target_img_, target_grid_, target_node_ = batch

        img[0]            = img_.astype(np.float32)
        target_img[0]     = target_img_.astype(np.float32)
        target_grid[0]    = target_grid_[None].astype(np.float32)
        target_node[0]    = target_node_[None].astype(np.float32)
        if config.inspection:
            inspection_logger.info("GetBatch: success")

    def grad(self, inputs, outputs_gradients):
        grad = [[disconnected_type(),]*4,]*3
        return grad

    def connection_pattern(self, node):
        return [[False, False, False, False],
                [False, False, False, False],
                [False, False, False, False]]

    def do_constant_folding(self, node):
        return False


class SkelGetBatchN(Node):
    """
    Dummy Node to be used in the split-function.

    Parameters
    ----------
    skel
    aux
    img_sh
    get_batch_kwargs
    scale_strenght
    name
    print_repr
    """

    def __init__(self, skel, aux, img_sh, get_batch_kwargs, scale_strenght=None,
                 name='skel_batch', print_repr=False):
        super(SkelGetBatchN, self).__init__((skel, aux), name, print_repr)
        self.skel = skel
        self.aux  = aux
        self.img_sh = img_sh
        self.get_batch_kwargs = get_batch_kwargs
        scale_strenght = scale_strenght if scale_strenght else 0.0
        self.scale_strenght = VariableParam(value=scale_strenght,
                                            name="scale_strenght",
                                            dtype=floatX,
                                            apply_train=False)

        self.params["scale_strenght"] = self.scale_strenght

    def _make_output(self):
        batch = SkelGetBatchOP()(self.skel.output, self.aux.output,
                                 self.scale_strenght, **self.get_batch_kwargs)
        self.output      = batch[0]
        self.target_img  = batch[1]
        self.target_grid = batch[2]
        self.target_node = batch[3]


    def _calc_shape(self):
        self.shape = TaggedShape(self.img_sh, 'b,f,z,y,x')

    def _calc_comp_cost(self):
        self.computational_cost = 0


def skelgetbatch_split(batch, img_sh, t_img_sh, t_grid_sh, t_node_sh,
                        name='skel_batch_split'):
    args = (img_sh, t_img_sh, t_grid_sh, t_node_sh)
    kwargs = dict(name=name)
    model_manager.current.register_split(batch, skelgetbatch_split, name, args, kwargs)
    # Split the various outputs of the batch
    img         = FromTensor(batch.output, img_sh, batch,
                             name='skel_img', print_repr=False)
    target_img  = FromTensor(batch.target_img, t_img_sh, batch,
                             name='skel_t_img', print_repr=False)
    target_grid = FromTensor(batch.target_grid, t_grid_sh, batch,
                             name='skel_t_grid', print_repr=False)
    target_node = FromTensor(batch.target_node, t_node_sh, batch,
                             name='skel_t_node', print_repr=False)

    return img, target_img, target_grid, target_node


def SkelGetBatch(skel, aux, img_sh, t_img_sh, t_grid_sh, t_node_sh,
                 get_batch_kwargs, scale_strenght=None, name='skel_batch'):
    # The SkelGetBatchN-Node must be created outside of Split because
    # Split is re-called at model restore and hence GetBatch would be create
    # Twice then
    get_batch_kwargs['t_grid_sh'] = t_grid_sh
    batch = SkelGetBatchN(skel, aux, img_sh, get_batch_kwargs,
                          scale_strenght=scale_strenght, name=name)
    return skelgetbatch_split(batch, img_sh, t_img_sh,  t_grid_sh, t_node_sh,
                              name=name+"_split")

class ScanN(Node):
    """
    WARNING: this node may only be used in conjuction with ``scansplit``
    because its ``output`` and ``shape`` attributes are lists which
    will confuse normal nodes. The split wraps the outputs in indivdual
    Nodes (FromTensor).

    Parameters
    ----------
    step_result: node/list(nodes)
        nodes that represent results of step function
    in_memory: node/list(nodes)
        nodes that inidcate at which place in the computational graph
        the memory is feed back into the step function. If ``out_memory``
        is not specified this must contain a node for *every* node in
        ``step_result`` because then the whole result will be fed back.
    out_memory: node/list(nodes)
        (optional) must be subset of ``step_result`` and of same length
        as ``in_memory``, tells which nodes of the result are fed back
        to ``in_memory``. If ``None``, all are fed back.
    in_iterate: node/list(nodes)
        nodes with a leading ``'r'`` axis to be iterated over (e.g.
        time series of shape [(30,r),(100,b),(50,f)]). In every step a slice
        from the first axis is consumed.
    in_iterate_0: node/list(nodes)
        nodes that consume a single slice of the ``in_iterate`` nodes.
        Part of "the inner function" of the scan loop in contrast to
        ``in_iterate``

    n_steps: int
    unroll_scan: bool
    last_only: bool
    name: str
    print_repr: bool
    """

    def __init__(self, step_result, in_memory, out_memory=None,
                 in_iterate=None, in_iterate_0=None, n_steps=None,
                 unroll_scan=True, last_only=False, name="scan", print_repr=True):
        step_result = utils.as_list(step_result)
        in_memory = utils.as_list(in_memory)
        out_memory = utils.as_list(out_memory)
        in_iterate = utils.as_list(in_iterate)
        in_iterate_0 = utils.as_list(in_iterate_0)

        if n_steps is None:
            assert in_iterate is not None
            if unroll_scan==True:
                n_steps = in_iterate[0].shape[0]
        else:
            if in_iterate is not None:
                assert n_steps == in_iterate[0].shape[0]

        if out_memory: assert len(in_memory)==len(out_memory)
        if in_iterate: assert in_iterate_0 and len(in_iterate)==len(in_iterate_0)
        if in_iterate_0: assert in_iterate
        if not out_memory: assert len(step_result)==len(in_memory)
        if out_memory:
            for o_m in out_memory: assert o_m in step_result
            #for o_m, i_m in zip(out_memory, in_memory): assert o_m.shape.shape == i_m.shape.shape
        else:
            for s_r, i_m in zip(step_result, in_memory): assert s_r.shape.shape == i_m.shape.shape

        parents = []
        for pl in [step_result, in_memory, in_iterate]:
            if pl is not None:
                parents.extend(pl)

        if out_memory:
            out_memory_sl = [step_result.index(s_r) for s_r in out_memory]
        else:
            out_memory_sl = list(range(len(step_result)))

        super(ScanN, self).__init__(parents, name, print_repr)
        self.step_result  = step_result
        self.in_memory    = in_memory
        self.out_memory   = out_memory
        self.out_memory_sl= out_memory_sl
        self.in_iterate   = in_iterate
        self.in_iterate_0 = in_iterate_0
        self._iterate     = in_iterate is not None

        self.n_steps      = n_steps
        self.unroll_scan  = unroll_scan
        self.last_only    = last_only
        self.output_names = ["scan_out_"+s_r.name for s_r in step_result]


    def _make_output(self):
        mem_hook = [i_m.output for i_m in self.in_memory]
        out_hook = [s_r.output for s_r in self.step_result]
        if self._iterate:
            it_hook = [i_o.output for i_o in self.in_iterate_0]

        if self.unroll_scan:
            out_accum = []
            # Doe one iteration as is
            new_out = [s_r.output for s_r in self.step_result]
            out_accum.append(new_out)
            # replace the memory input and iterate in every step
            for t in range(1, self.n_steps):
                new_mem = [new_out[i] for i in self.out_memory_sl]
                replacements = dict(zip(mem_hook, new_mem))
                if self._iterate:
                    for i_h, i_i in zip(it_hook, self.in_iterate):
                        replacements[i_h] = i_i.output[t]

                new_out = theano.clone(out_hook, replace=replacements)
                out_accum.append(new_out)
                # for k,v in replacements.items():
                #      print(" Replace ",k.auto_name,k," by ",v.auto_name,v)
                # print('---')

            if self.last_only:
                self.output = new_out
            else:
                self.output = [T.stack(r_accum, axis=0) for r_accum in
                               zip(*out_accum)]

        else:
            if self.n_steps:
                logger.warning("If the number of steps is known, "
                               "you should use 'unroll_scan'!")
            assert self._iterate
            l_it = len(it_hook)
            l_out = len(out_hook)
            l_mem = len(mem_hook)
            mem_map = []
            for s_r in self.step_result:
                try:
                    ix = self.out_memory.index(s_r)
                except:
                    ix = None
                mem_map.append(ix)

            outputs_info = []
            for i, s_r in enumerate(out_hook):
                if mem_map[i] is not None:
                    outputs_info.append(dict(initial=out_hook[mem_map[i]], taps=[-1]))
                else:
                    outputs_info.append(None)

            def step(*args):
                # args: input0 slices, output slices, non_seq
                it_slice     = args[0:l_it]
                mem_new_hook = args[l_it:l_it+l_mem]
                tmp = args[l_it+l_mem:]
                out_hook__   = tmp[0:l_out]
                it_hook__    = tmp[l_out:l_it+l_out]
                mem_hook__   = tmp[l_it+l_out:]
                replacements__ = dict()
                for i, m_i, in enumerate(out_hook__):
                    if mem_map[i] is not None:
                        ix = mem_map[i]
                        replacements__[mem_hook__[ix]] = mem_new_hook[i]

                for i_h, i_i in zip(it_hook__, it_slice): # consume the in_iterate_series slice
                    replacements__[i_h] = i_i

                new_outputs = theano.clone(out_hook__, replace=replacements__)
                updates__ = dict()
                #condition = theano.scan_module.until(False)
                return new_outputs, updates__ #, condition

            results, updates = theano.scan(step,
                                           sequences=[i_i.output[1:] for i_i in self.in_iterate],
                                           outputs_info=outputs_info,
                                           non_sequences=out_hook+it_hook+mem_hook)
            if self.last_only:
                self.output = [r[-1] for r in results]
            else:
                l = []
                results = utils.as_list(results)
                for r_0, r_accum in zip(out_hook, results):
                    pattern = ['x'] + list(range(r_0.ndim))
                    l.append(T.concatenate([r_0.dimshuffle(pattern), r_accum], axis=0))

                self.output = l



    def _calc_shape(self):
        if self.n_steps is None:
            n = self.in_iterate[0].shape[0]
        else:
            n = self.n_steps
        if self.last_only:
            self.shape = [s_r.shape.copy() for s_r in self.step_result]
        else:
            self.shape = [s_r.shape.addaxis(0, n, 'r') for s_r in self.step_result]


    def _calc_comp_cost(self):
        if self.n_steps is None:
            n = self.in_iterate[0].shape[0]
        else:
            n = self.n_steps

        self.computational_cost = (n-1) * self.step_result[0].all_computational_cost

def scansplit(scan, name='scan_split'):
    args = ()
    kwargs = dict(name=name)
    model_manager.current.register_split(scan, scansplit, name, args,
                                         kwargs)

    outs = []
    for out, out_sh, out_name in zip(scan.output, scan.shape, scan.output_names):
        outs.append(FromTensor(out, out_sh, scan,
                     name=out_name, print_repr=False))

    return outs


def Scan(step_result, in_memory, out_memory=None,
         in_iterate=None, in_iterate_0=None, n_steps=None,
         unroll_scan=True, last_only=False, name="scan", print_repr=True):
    """
    Parameters
    ----------
    step_result: node/list(nodes)
        nodes that represent results of step function
    in_memory: node/list(nodes)
        nodes that inidcate at which place in the computational graph
        the memory is feed back into the step function. If ``out_memory``
        is not specified this must contain a node for *every* node in
        ``step_result`` because then the whole result will be fed back.
    out_memory: node/list(nodes)
        (optional) must be subset of ``step_result`` and of same length
        as ``in_memory``, tells which nodes of the result are fed back
        to ``in_memory``. If ``None``, all are fed back.
    in_iterate: node/list(nodes)
        nodes with a leading ``'r'`` axis to be iterated over (e.g.
        time series of shape [(30,r),(100,b),(50,f)]). In every step a slice
        from the first axis is consumed.
    in_iterate_0: node/list(nodes)
        nodes that consume a single slice of the ``in_iterate`` nodes.
        Part of "the inner function" of the scan loop in contrast to
        ``in_iterate``

    n_steps: int
    unroll_scan: bool
    last_only: bool
    name: str
    print_repr: bool

    Returns
    -------

    A node for every node in ``step_result`` which either contains the last
    state or the series of states - then it has a leading ``'r'`` axis.
    """
    scan = ScanN(step_result, in_memory, out_memory, in_iterate, in_iterate_0,
                 n_steps, unroll_scan=unroll_scan, last_only=last_only,
                 name=name, print_repr=print_repr)

    outputs = scansplit(scan, name='scan_split')
    if len(outputs)==1:
        outputs = outputs[0]

    return outputs


class Reshape(Node):
    """
    Reshape node.

    Parameters
    ----------
    parent
    shape
    tags
    strides
    fov
    name
    print_repr
    """
    def __init__(self, parent, shape, tags=None, strides=None, fov=None, name="reshape", print_repr=True):

        super(Reshape, self).__init__(parent, name, print_repr)
        self.parent   = parent
        if isinstance(shape, TaggedShape):
            self._shape = shape
        else:
            if tags is None:
                raise ValueError("Tags argument must not be None if "
                                 "shape is not a TaggedShape")
            self._shape = TaggedShape(shape, tags, strides, fov=fov)

        if self._shape.stripbatch_prod != parent.shape.stripbatch_prod:
            raise ValueError("Cannot reshap %s to %s" %(parent.shape, self._shape))

    def _make_output(self):
        new_sh = [x if x is not None else -1 for x in self._shape]
        self.output = self.parent.output.reshape(new_sh)

    def _calc_shape(self):
        self.shape = self._shape

    def _calc_comp_cost(self):
        self.computational_cost = 0


_SkeletonGetBatch = SkelGetBatchN
_skeletongetbatchsplit = skelgetbatch_split
SkeletonLossRec = SkelLossRec
SkeletonPrior = SkelPrior
_Scan = ScanN
_scansplit = scansplit
_SkeletonLoss = SkelLoss

if __name__=="__main__":
    # test wrapping of generic type into graph
    class DummySkel(object):
        def __init__(self):
            self.val = 3.141
            self.grad = np.array([1, 3, 5], dtype=np.float32)

        def get_loss(self, predicted):
            # print("Im here %s" %predicted)
            return self.val, self.grad * predicted

    skel_var = gof.type.Generic()()
    x = T.TensorType(dtype='float32', broadcastable=(False,))()
    w = theano.shared(np.eye(3, dtype=np.float32))
    pred_var = theano.dot(w, x)

    loss_var, loss_grad = SkelLossOP()(pred_var, skel_var)

    f = theano.function([x, skel_var], [loss_var])
    grad_var = theano.grad(loss_var, [w, ])
    f_grad = theano.function([x, skel_var], grad_var + [loss_var, pred_var])

    test_vals = np.array([1, 2, 3], dtype=np.float32)
    skel = DummySkel()

    print(f(test_vals, skel))
    print(f_grad(test_vals, skel))
    theano.printing.pydotprint(f_grad, '/tmp/step_result.svg')

    ###########################################################################

    from elektronn2 import neuromancer
    import theano
    act = 'tanh'
    data = neuromancer.Input((1, 20), 'b,f', name='data')
    mem_0 = neuromancer.Input((1, 120), 'b,f', name='mem')
    mlp1 = neuromancer.Dot(data, 120, activation_func=act)
    join = neuromancer.Concat([mlp1, mem_0])
    out = neuromancer.Dot(join, 120, activation_func=act)
    out2 = neuromancer.Dot(out, 13, activation_func='lin')
    # recurrent    = neuromancer.Scan(out, in_memory=mem_0, n_steps=10)
    recurrent, out2r = neuromancer.Scan([out, out2], out_memory=out,
                                        in_memory=mem_0, n_steps=7)
    loss = neuromancer.AggregateLoss(recurrent)
    loss = neuromancer.AggregateLoss(out2r)

    grad = theano.grad(loss.output, loss.all_trainable_params.values(),
                       disconnected_inputs='warn')

    recurrent()
    out2r()
    print(len(list(filter(lambda x: x.op.__class__.__name__=="Dot22",
                    recurrent._output_func.func.maker.fgraph.apply_nodes))))

    theano.printing.pydotprint(recurrent._output_func.func,
                               outfile="/tmp/test-comp.png",
                               var_with_name_simple=True)
    theano.printing.pydotprint(out2r._output_func.func,
                               outfile="/tmp/test2-comp.png",
                               var_with_name_simple=True)
    fn = theano.function(out.input_tensors, [recurrent.output, out2r.output])
    x = np.random.rand(1, 20).astype(np.float32)
    m = np.random.rand(1, 120).astype(np.float32)
    y = recurrent(x, m)
    z = out2r(x, m)
    grad_fn = theano.function(out.input_tensors, grad)
    g = grad_fn(x, m)



# -*- coding: utf-8 -*-

# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, \
    super, zip

import logging
import numpy as np


__all__ = ["cnncalculator", "get_cloesest_valid_patch_size",
           "get_valid_patch_sizes"]

logger = logging.getLogger('elektronn2log')


class _Layer(object):
    def __init__(self, patch_size, filter=1, pool=1, stride=1, mfp=True):
        self.field = None
        self.overlap = None

        self.out = patch_size - filter + 1
        if self.out <= 0:
            raise ValueError('CNN has no output for Layer with patch_size',
                             patch_size)
        self.stride = stride
        rest = self.out % pool
        self.pool_out = self.out // pool
        if pool > 1:
            if mfp and rest!=1:
                raise ValueError('mfp fails for Layer with patch_size',
                                 patch_size)
            elif not mfp and rest > 0:
                raise ValueError('Uneven Pools for Layer with patch_size',
                                 patch_size)

    def setfield(self, field):
        self.field = field
        self.overlap = field - self.stride


class _Cnncalculator(object):
    def __init__(self, filters, poolings, desired_patch_size, mfp,
                 force_center, desired_output):
        self.layers = None
        self.pool_out = None
        self.out = None
        self.stride = None
        self.mfp = mfp
        self.fields = self.getFields(filters, poolings)
        fow = self.fields[-1]
        if fow % 2==0:
            if force_center:
                raise ValueError('Receptive Fields are not '
                                 'centered with field of view (%i)' % fow)
            else:
                logger.warning('WARNING: Receptive Fields are not centered '
                               'with even field of view (%i)' % fow)
        self.offset = float(fow) / 2

        valid_patch_sizes = []
        valid_outputs = []
        for inp in range(2, 5000):
            try:
                self.calclayers(inp, filters, poolings, mfp)
                valid_patch_sizes.append(inp)
                valid_outputs.append(self.out[-1])
            except:
                pass

        if desired_output is not None:
            if desired_output in valid_outputs:
                i = valid_outputs.index(desired_output)
                patch_size = valid_patch_sizes[i]
            else:
                valid_outputs = np.array(valid_outputs)
                patch_size = valid_outputs[valid_outputs <= desired_output][-1]
                logger.info("Info: output size requires patch_size>5000, "
                            "next smaller output (%i) is used" % patch_size)
                valid_outputs = list(valid_outputs)

            # patch_size corresponding to that output
            i = valid_outputs.index(desired_output)
            patch_size = valid_patch_sizes[i]

        elif desired_patch_size in valid_patch_sizes:
            patch_size = desired_patch_size
        elif desired_patch_size is None:
            patch_size = valid_patch_sizes[-1]
        else:
            valid_patch_sizes = np.array(valid_patch_sizes)
            if desired_patch_size < valid_patch_sizes[0]:
                patch_size = valid_patch_sizes[0]
                logger.info("patch_size (%i) changed to (%i) "
                            "(size too small)" % (
                                desired_patch_size, patch_size))
            else:
                patch_size = valid_patch_sizes[
                    valid_patch_sizes <= desired_patch_size][-1]
                logger.info("patch_size (%i) changed to (%i) (size not "
                            "possible)" % (desired_patch_size, patch_size))
                valid_patch_sizes = list(valid_patch_sizes)

        self.valid_patch_sizes = valid_patch_sizes
        self.calclayers(patch_size, filters, poolings, mfp)
        self.patch_size = patch_size
        self.pred_stride = self.layers[-1].stride
        for lay, field in zip(self.layers, self.fields):
            lay.setfield(field)

        self.overlap = [l.overlap for l in self.layers]

    def calclayers(self, patch_size, filters, poolings, mfp):
        stride = poolings[0]
        lay0 = _Layer(patch_size, filters[0], poolings[0], stride, mfp=mfp[0])
        self.layers = [lay0,]
        for i in range(1, len(filters)):
            stride = np.multiply(stride, poolings[i])
            lay = _Layer(self.layers[i - 1].pool_out, filters[i], poolings[i],
                         stride, mfp[i])
            self.layers.append(lay)

        self.pool_out = [l.pool_out for l in self.layers]
        self.out = [l.out for l in self.layers]
        self.stride = [l.stride for l in self.layers]

    def __repr__(self):
        if not isinstance(self.pool_out[0], list):
            ls = self.pool_out[::-1]
        else:
            ls = list(zip(*self.pool_out))[::-1]
        if not isinstance(self.out[0], list):
            out = self.out[::-1]
        else:
            out = list(zip(*self.out))[::-1]
        if not isinstance(self.fields[0], list):
            fields = self.fields[::-1]
        else:
            fields = list(zip(*self.fields))[::-1]
        if not isinstance(self.stride[0], list):
            stride = self.stride[::-1]
        else:
            stride = list(zip(*self.stride))[::-1]
        if not isinstance(self.overlap[0], list):
            overlap = self.overlap[::-1]
        else:
            overlap = list(zip(*self.overlap))[::-1]

        s = "patch_size: " + repr(
            self.patch_size) + "\nLayer/Fragment sizes:\t" + repr(ls)\
            + "\nUnpooled Layer sizes:\t" + repr(out)\
            + "\nReceptive fields:\t" + repr(fields)\
            + "\nStrides:\t\t" + repr(stride)\
            + "\nOverlap:\t\t" + repr(overlap) + "\nOffset:\t\t"\
            + repr(self.offset) + "\nIf offset is non-int: output neurons " \
            + "lie centered on patch_size neurons,they have an odd FOV\n"
        return s

    @staticmethod
    def getFields(filter, pool):
        def recFields_helper(filter, pool):
            rf = [None] * (len(filter) + 1)
            rf[-1] = 1
            for i in range(len(filter), 0, -1):
                rf[i - 1] = rf[i] * pool[i - 1] + filter[i - 1] - 1
            return rf[0]

        fields = []
        for i in range(1, len(filter) + 1):
            fields.append(recFields_helper(filter[:i], pool[:i]))

        return fields


class _Multi_cnncalculator(_Cnncalculator):
    """ Adaptor Class to unify multiple CNNCalculators"""

    def __init__(self, calcs):
        self.fields = []
        self.offset = []
        self.valid_patch_sizes = []
        self.patch_size = []
        self.pred_stride = []
        self.stride = []
        self.pool_out = []
        self.out = []
        self.overlap = []
        for c in calcs:
            self.fields.append(c.fields)
            self.offset.append(c.offset)
            self.valid_patch_sizes.append(c.valid_patch_sizes)
            self.patch_size.append(c.patch_size)
            self.pred_stride.append(c.pred_stride)
            self.overlap.append(c.overlap)
            self.pool_out.append(c.pool_out)
            self.out.append(c.out)
            self.stride.append(c.stride)


def cnncalculator(filters, poolings, desired_patch_size=None, mfp=False,
                  force_center=False, desired_output=None, ndim=1):
    """
    Helper to calculate CNN architectures

    This is a *function*, but it returns an *object* that has various
    architecture values as attributes.
    Useful is also to simply print 'd' as in the example.

    Parameters
    ----------

    filters: list
      Filter shapes (for anisotropic filters the shapes are again a list)
    poolings: list
      Pooling factors
    desired_patch_size: int or list of int
      Desired patch_size size(s). If ``None`` a range of suggestions can be
      found in the attribute ``valid_patch_sizes``
    mfp: list of int/{0,1}
      Whether to apply Max-Fragment-Pooling in this Layer and check
      compliance with max-fragment-pooling
      (requires other patch_size sizes than normal pooling)
    force_center: Bool
      Check if output neurons/pixel lie at center of patch_size
      neurons/pixel (and not in between)
    desired_output: int or list of int
      Alternative to ``desired_patch_size``
    ndim: int
      Dimensionality of CNN

    Examples
    --------

    Calculation for anisotropic "flat" 3d CNN with mfp in the first layers only::

      >>> desired_patch_size   = [211, 211, 20]
      >>> filters         = [[6,6,1], [4,4,4], [2,2,2], [1,1,1]]
      >>> pool            = [[2,2,1], [2,2,2], [2,2,2], [1,1,1]]
      >>> mfp             = [1,        1,       0,       0,   ]
      >>> ndim=3
      >>> d = cnncalculator(filters, pool, desired_patch_size, mfp=mfp, force_center=True, desired_output=None, ndim=ndim)
      Info: patch_size (211) changed to (210) (size not possible)
      Info: patch_size (211) changed to (210) (size not possible)
      Info: patch_size (20) changed to (22) (size too small)
      >>> print(d)
      patch_size: [210, 210, 22]
      Layer/Fragment sizes:	[[102, 49, 24, 24], [102, 49, 24, 24], [22, 9, 4, 4]]
      Unpooled Layer sizes:	[[205, 99, 48, 24], [205, 99, 48, 24], [22, 19, 8, 4]]
      Receptive fields:	[[7, 15, 23, 23], [7, 15, 23, 23], [1, 5, 9, 9]]
      Strides:		[[2, 4, 8, 8], [2, 4, 8, 8], [1, 2, 4, 4]]
      Overlap:		[[5, 11, 15, 15], [5, 11, 15, 15], [0, 3, 5, 5]]
      Offset:		[11.5, 11.5, 4.5].
          If offset is non-int: floor(offset).
          Select labels from within img[offset-x:offset+x]
          (non-int means, output neurons lie centered on patch_size neurons,
          i.e. they have an odd field of view)
    """  # TODO: The example is still ELEKTRONN1-specific. Check the whole function for incompatibilities.

    assert len(poolings)==len(filters)

    if mfp is False:
        mfp = [False, ] * len(filters)

    if ndim==1:  # not hasattr(filters[0], '__len__') :
        return _Cnncalculator(filters, poolings, desired_patch_size, mfp,
                              force_center, desired_output)
    else:
        if desired_patch_size is None:
            desired_patch_size = (None,) * ndim
        elif not hasattr(desired_patch_size, '__len__'):
            desired_patch_size = (desired_patch_size,) * ndim
        if desired_output is None:
            desired_output = (None,) * ndim
        elif not hasattr(desired_output, '__len__'):
            desired_output = (desired_output,) * ndim
        if not hasattr(poolings[0], '__len__'):
            poolings = [[p, ] * ndim for p in poolings]
        if not hasattr(filters[0], '__len__'):
            filters = [[f, ] * ndim for f in filters]
        if not hasattr(mfp[0], '__len__'):
            mfp = [[m, ] * ndim for m in mfp]

        assert len(mfp)==len(filters)

        filters = [list(l) for l in zip(*filters)]
        poolings = [list(l) for l in zip(*poolings)]
        mfp = [list(l) for l in zip(*mfp)]

        calcs = []
        for f, p, d, o, mfp in zip(filters, poolings, desired_patch_size,
                                   desired_output, mfp):
            c = _Cnncalculator(f, p, d, mfp, force_center, o)
            calcs.append(c)

        return _Multi_cnncalculator(calcs)


def get_valid_patch_sizes(filters, poolings, desired_patch_size=100,mfp=False,
                          ndim=1):
    calc = cnncalculator(filters, poolings, desired_patch_size, mfp=mfp,
                         ndim=ndim)
    return calc.valid_patch_sizes


def get_cloesest_valid_patch_size(filters, poolings, desired_patch_size=100,
                                  mfp=False, ndim=1):
    calc = cnncalculator(filters, poolings, desired_patch_size, mfp=mfp,
                         ndim=ndim)
    return calc.patch_size


if __name__=="__main__":
    logger.debug("Testing cnncalculator")
    desired_patch_size = 200
    mfp = False
    filters = [6, 5, 4, 4, 1, 4, 4, 4, 4, 2]
    pool = [1, 2, 2, 1, 1, 1, 1, 1, 1, 1]

    #    filters         = [1,1,1,1, 4,3,3,2, 2,1]
    #    pool            = [1,1,1,1, 2,1,1,1, 1,1]

    ndim = 1
    d = cnncalculator(filters, pool, desired_patch_size, mfp=mfp,
                      force_center=True, desired_output=None, ndim=ndim)
    print(d)



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved
# This Code is adapted from Sven Dorkenwald

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, \
    super, zip

import sys
import subprocess
import time


def initgpu(gpu):
    if gpu is None:
        gpu = 'none'
    no_gpu = ['none', 'None']
    import theano.sandbox.cuda

    if theano.sandbox.cuda.cuda_available:
        if isinstance(gpu, str) and gpu.lower() == 'auto':
            gpu = int(get_free_gpu())
            print("Automatically assigned free GPU %i" % (gpu,))

        if gpu in no_gpu and gpu != 0:
            pass
        else:
            try:
                theano.sandbox.cuda.use("gpu" + str(gpu))
                print("Initialising GPU to %s" % gpu)
            except:
                sys.excepthook(*sys.exc_info())
                print("Failed to init GPU, argument not understood.")

    else:
        if gpu in no_gpu and gpu != 0:
            pass
        else:
            print("'--gpu' argument is not 'none' but CUDA is not available. "
                  "Falling back to CPU.")


def _check_if_gpu_is_free(nb_gpu):
    process_output = subprocess.Popen('nvidia-smi -i %d -q -d PIDS' % nb_gpu,
                                      stdout=subprocess.PIPE,
                                      shell=True).communicate()[0]
    if b"Process ID" in process_output and b"Used GPU Memory" in process_output:
        return 0
    else:
        return 1


def _get_number_gpus():
    process_output = subprocess.Popen('nvidia-smi -L', stdout=subprocess.PIPE,
                                      shell=True).communicate()[0].decode()
    nb_gpus = 0
    while True:
        if "GPU %d" % nb_gpus in process_output:
            nb_gpus += 1
        else:
            break
    return nb_gpus


def get_free_gpu(wait=0, nb_gpus=-1):
    import theano.sandbox.cuda

    if not theano.sandbox.cuda.cuda_available:
        print("Cannot get free gpu. Cuda not available on this "
              "host")
        return -1
    if nb_gpus==-1:
        nb_gpus = _get_number_gpus()
    while True:
        for nb_gpu in range(nb_gpus):
            if _check_if_gpu_is_free(nb_gpu)==1:
                return nb_gpu
        if wait > 0:
            time.sleep(2)
        else:
            return -1



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved
from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip


import numpy as np
import theano
from theano import gof
import theano.tensor as T
from theano.gradient import disconnected_type

from . import malis_utils

__all__ = ['malis_weights']

class MalisWeights(theano.Op):
    """
    Computes MALIS loss weights
    
    Roughly speaking the malis weights quantify the impact of an edge in
    the predicted affinity graph on the resulting segmentation.
      
    Parameters
    ----------
    affinity_pred: 4d np.ndarray float32
        Affinity graph of shape (#edges, x, y, z)
        1: connected, 0: disconnected        
    affinity_gt: 4d np.ndarray int16
        Affinity graph of shape (#edges, x, y, z)
        1: connected, 0: disconnected 
    seg_gt: 3d np.ndarray, int (any precision)
        Volume of segmentation IDs        
    nhood: 2d np.ndarray, int
        Neighbourhood pattern specifying the edges in the affinity graph
        Shape: (#edges, ndim)
        nhood[i] contains the displacement coordinates of edge i
        The number and order of edges is arbitrary
    unrestrict_neg: Bool
        Use this to relax the restriction on neg_counts. The restriction
        modifies the edge weights for before calculating the negative counts
        as: ``edge_weights_neg = np.maximum(affinity_pred, affinity_gt)``
        If unrestricted the predictions are used directly.
        
    Returns
    -------
      
    pos_counts: 4d np.ndarray int32
      Impact counts for edges that should be 1 (connect)      
    neg_counts: 4d np.ndarray int32
      Impact counts for edges that should be 0 (disconnect)  
      
      
    Outline
    -------
      
    - Computes for all pixel-pairs the MaxiMin-Affinity
    - Separately for pixel-pairs that should/should not be connected
    - Every time an affinity prediction is a MaxiMin-Affinity its weight is
      incremented by one in the output matrix (in different slices depending
      on whether that that pair should/should not be connected)
    """
    __props__ = ()

    def make_node(self, *inputs):
        inputs = list(inputs)
        if len(inputs)!=5:
            raise ValueError("MalisOp takes 5 inputs: \
          affinity_pred, affinity_gt, seg_gt, nhood, unrestrict_neg")
        
        inputs = list(map(T.as_tensor_variable, inputs))
        
        affinity_pred, affinity_gt, seg_gt, nhood = inputs[:4]
        if affinity_pred.ndim!=4:
            raise ValueError("affinity_pred must be convertible to a\
          TensorVariable of dimensionality 4. This one has ndim=%i" \
                             %affinity_pred.ndim)

        if affinity_gt.ndim!=4:
            raise ValueError("affinity_gt must be convertible to a\
          TensorVariable of dimensionality 4. This one has ndim=%i" \
                             %affinity_gt.ndim)

        if seg_gt.ndim!=3:
            raise ValueError("seg_gt must be convertible to a\
          TensorVariable of dimensionality 3. This one has ndim=%i" \
                             %seg_gt.ndim)

        if nhood.ndim!=2:
            raise ValueError("nhood must be convertible to a\
          TensorVariable of dimensionality 2. This one has ndim=%i" \
                             %nhood.ndim)
                                     
        malis_weights_pos = T.TensorType(
            dtype='uint64',
            broadcastable=(False,)*4)()

        malis_weights_neg = T.TensorType(
            dtype='uint64',
            broadcastable=(False,)*4)()

        outputs = [malis_weights_pos, malis_weights_neg]

        return gof.Apply(self, inputs, outputs)

    def perform(self, node, inputs, output_storage):
        affinity_pred, affinity_gt, seg_gt, nhood, unrestrict_neg = inputs
        pos, neg = output_storage
        pos[0], neg[0] = malis_utils.malis_weights(affinity_pred,
                                                   affinity_gt,
                                                   seg_gt,
                                                   nhood,
                                                   unrestrict_neg)

    def grad(self, inputs, outputs_gradients):
        # The gradient of all outputs is 0 w.r.t to all inputs
        return [disconnected_type(),]*5

    def connection_pattern(self, node):
        # The gradient of all outputs is 0 w.r.t to all inputs
        return [[False, False],]*5

def malis_weights(affinity_pred, affinity_gt, seg_gt, nhood, unrestrict_neg=False):
    """
    Computes MALIS loss weights
    
    Roughly speaking the malis weights quantify the impact of an edge in
    the predicted affinity graph on the resulting segmentation.
      
    Parameters
    ----------
    affinity_pred: 4d np.ndarray float32
        Affinity graph of shape (#edges, x, y, z)
        1: connected, 0: disconnected        
    affinity_gt: 4d np.ndarray int16
        Affinity graph of shape (#edges, x, y, z)
        1: connected, 0: disconnected 
    seg_gt: 3d np.ndarray, int (any precision)
        Volume of segmentation IDs        
    nhood: 2d np.ndarray, int
        Neighbourhood pattern specifying the edges in the affinity graph
        Shape: (#edges, ndim)
        nhood[i] contains the displacement coordinates of edge i
        The number and order of edges is arbitrary
    unrestrict_neg: Bool
        Use this to relax the restriction on neg_counts. The restriction
        modifies the edge weights for before calculating the negative counts
        as: ``edge_weights_neg = np.maximum(affinity_pred, affinity_gt)``
        If unrestricted the predictions are used directly.
        
    Returns
    -------
      
    pos_counts: 4d np.ndarray int32
      Impact counts for edges that should be 1 (connect)      
    neg_counts: 4d np.ndarray int32
      Impact counts for edges that should be 0 (disconnect)  
      
      
    Outline
    -------
      
    - Computes for all pixel-pairs the MaxiMin-Affinity
    - Separately for pixel-pairs that should/should not be connected
    - Every time an affinity prediction is a MaxiMin-Affinity its weight is
      incremented by one in the output matrix (in different slices depending
      on whether that that pair should/should not be connected)
    """
    rest = 1 if unrestrict_neg else 0 # Theano cannot bool        
    return MalisWeights()(affinity_pred, affinity_gt, seg_gt, nhood, rest)



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, \
    super, zip

__all__ = ['CircularBuffer', 'AccumulationArray', 'DynamicKDT', 'KDT',
           'pickleload', 'picklesave', 'h5load', 'h5save', 'pretty_string_ops',
           'import_variable_from_file', 'timeit', 'Timer',
           'cache', 'pretty_string_time', 'unique_rows',
           'get_free_cpu_count', 'parallel_accum', 'makeversiondir', 'as_list']

from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip

import os
import re
import time
import logging
from functools import reduce

try:
    import cPickle as pkl
except:
    import pickle as pkl


import numba
import psutil
from multiprocessing import Pool
import gzip
import h5py
import numpy as np
import sys
import importlib
import functools
from scipy.spatial.distance import cdist
import sklearn
from sklearn.neighbors import NearestNeighbors as NearestNeighbors_


# TODO: Why are there two __all__ assignments? Delete the first one?
__all__ = ['get_free_cpu_count', 'parallel_accum',
           'timeit', 'cache', 'CircularBuffer', 'AccumulationArray', 'KDT',
           'DynamicKDT', 'import_variable_from_file', 'pickleload', 'picklesave',
           'h5save', 'h5load', 'pretty_string_ops', 'pretty_string_time',
           'makeversiondir', 'Timer', 'unique_rows', 'as_list']

logger = logging.getLogger('elektronn2log')


def get_free_cpu_count():
    m = psutil.cpu_count()
    if m<=2:
        return 1
    else:
        load = float(psutil.cpu_percent(interval=0.4)) / 100
        free = 1 - load
        n = min(max(3, m * free * 0.8), m - 1)
        return int(n)


def parallel_accum(func, n_ret, var_args, const_args, proc=-1, debug=False):
    if proc==-1:
        proc = get_free_cpu_count()
    args = []
    for a in var_args:
        try:
            v = tuple(a)
        except TypeError:
            v = (a,)

        arg = v + tuple(const_args)
        args.append(arg)

    assert len(args)>0
    if not debug:
        p = Pool(proc)
        try:
            ret = p.map(func, args)
            p.close()
            p.join()
        except KeyboardInterrupt:
            p.terminate()
            p.join()
            raise KeyboardInterrupt
    else:
        ret = map(func, args)

    accums = [list() for i in range(n_ret)]
    for tmp in ret:
        if n_ret==1:
            try:
                accums[0].extend(tmp)
            except TypeError:
                accums[0].append(tmp)
        else:
            for i in range(n_ret):
                x = tmp[i]
                try:
                    accums[i].extend(x)
                except TypeError:
                    accums[i].append(x)

    if n_ret==1:
        return accums[0]
    else:
        return tuple(accums)


### Decorator Collection ###

class DecoratorBase(object):
    """
    If used as
    ``@DecoratorBase``
    this initialiser receives only the function to be wrapped (no wrapper args)
    Then ``__call__`` receives the arguments for the underlying function.

    Alternatively, if used as
    ``@DecoratorBase(wrapper_print=True, n_times=10)``
    this initialiser receives wrapper args, the function is passed to ``__call__``
    and ``__call__`` returns a wrapped function.

    This base class completely ignores all wrapper arguments.
    """

    def __init__(self, *args, **kwargs):
        self.func = None
        self.dec_args = None
        self.dec_kwargs = None
        if len(args)==1 and not len(kwargs):
            assert hasattr(args[0], '__call__')
            func = args[0]
            self.func = func
            self.__call__.__func__.__doc__ = func.__doc__
            self.__call__.__func__.__name__ = func.__name__
        else:
            self.dec_args = args
            self.dec_kwargs = kwargs

    def __call__(self, *args, **kwargs):
        # The decorator was initialised with the func, it now has apply the decoration itself
        if not self.func is None:
            # do something with args
            ret = self.func(*args, **kwargs)
            # do something with kwargs
            return ret

        # The decorator was initialised with args, it now returns a wrapped function
        elif len(args)==1 and not len(kwargs):
            assert hasattr(args[0], '__call__')
            func = args[0]

            @functools.wraps(func)
            def decorated(*args0, **kwargs0):
                # do something with args0, read the decorator arguments
                # print(self.dec_args)
                # print(self.dec_kwargs)
                ret = func(*args0, **kwargs0)
                # do something with ret
                return ret

            return decorated
        else:
            raise ValueError()


class timeit(DecoratorBase):
    def __call__(self, *args, **kwargs):
        # The nor args for the decorator --> n=1
        if not self.func is None:
            t0 = time.time()
            ret = self.func(*args, **kwargs)
            t = time.time() - t0
            print("Function <%s> took %.5g s" % (self.func.__name__, t))
            return ret

        # The decorator was initialised with args, it now returns a wrapped function
        elif len(args)==1 and not len(kwargs):
            assert hasattr(args[0], '__call__')
            func = args[0]
            n = self.dec_kwargs.get('n', 1)

            @functools.wraps(func)
            def decorated(*args0, **kwargs0):
                t0 = time.time()
                if n>1:
                    for i in range(n - 1):
                        func(*args0, **kwargs0)

                ret = func(*args0, **kwargs0)
                t = time.time() - t0
                print("Function <%s> took %.5g s averaged over %i execs" % (
                    func.__name__, t / n, n))

                return ret

            return decorated

        else:
            raise ValueError()


class cache(DecoratorBase):
    def __init__(self, *args, **kwargs):
        super(cache, self).__init__(*args, **kwargs)
        self.memo = {}
        self.default = None

    @staticmethod
    def hash_args(args):
        tmp = []
        for arg in args:
            if isinstance(arg, np.ndarray):
                tmp.append(hash(arg.tostring()))
            elif isinstance(arg, (list, tuple)):
                tmp.append(reduce(lambda x, y: x + hash(y), arg, 0))
            else:
                tmp.append(hash(arg))

        return reduce(lambda x, y: x + y, tmp, 0)

    def __call__(self, *args, **kwargs):
        # The nor args for the decorator --> n=1
        if not self.func is None:
            if len(args)==0 and len(kwargs)==0:
                if self.default is None:
                    self.default = self()
                return self.default()
            else:
                key1 = self.hash_args(args)
                key2 = self.hash_args(kwargs.values())
                key = key1 + key2
                if not key in self.memo:
                    self.memo[key] = self.func(*args, **kwargs)
                return self.memo[key]

        # The decorator was initialised with args, it now returns a wrapped function
        elif len(args)==1 and not len(kwargs):
            assert hasattr(args[0], '__call__')
            func = args[0]

            @functools.wraps(func)
            def decorated(*args0, **kwargs0):
                if len(args0)==0 and len(kwargs0)==0:
                    if self.default is None:
                        self.default = self()
                    return self.default()
                else:
                    key1 = self.hash_args(args0)
                    key2 = self.hash_args(kwargs0.values())
                    key = key1 + key2
                    if not key in self.memo:
                        self.memo[key] = func(*args0, **kwargs0)
                    return self.memo[key]

            return decorated

        else:
            raise ValueError()


### Custom Data Structures ###

class CircularBuffer(object):
    def __init__(self, buffer_len):
        self.length = 0
        self._buffer = np.zeros(buffer_len)

    def append(self, data):
        self._buffer = np.roll(self._buffer, 1)
        self._buffer[0] = data
        self.length += 1

    @property
    def data(self):
        return self._buffer[:self.length]

    def mean(self):
        if self.length:
            return self.data.mean()
        else:
            return 0.0

    def setvals(self, val):
        self._buffer[:] = val

    def __len__(self):
        return self.length

    def __getitem__(self, slc):
        return self._buffer[self.length][slc]

    def __repr__(self):
        return repr(self.data)


class AccumulationArray(object):
    def __init__(self, right_shape=(), dtype=np.float32, n_init=100, data=None,
                 ema_factor=0.95):
        if isinstance(dtype, dict) and right_shape!=():
            raise ValueError("If dict is used as dtype, right shape must be"
                             "unchanged (i.e it is 1d)")

        if data is not None and len(data):
            n_init += len(data)
            right_shape = data.shape[1:]
            dtype = data.dtype

        self._n_init = n_init
        if isinstance(right_shape, int):
            self._right_shape = (right_shape,)
        else:
            self._right_shape = tuple(right_shape)
        self.dtype = dtype
        self.length = 0
        self._buffer = self._alloc(n_init)
        self._min = +np.inf
        self._max = -np.inf
        self._sum = 0
        self._ema = None
        self._ema_factor = ema_factor

        if data is not None and len(data):
            self.length = len(data)
            self._buffer[:self.length] = data
            self._min = data.min(0)
            self._max = data.max(0)
            self._sum = data.sum(0)

    def __repr__(self):
        return repr(self.data)

    def _alloc(self, n):
        if isinstance(self._right_shape, (tuple, list, np.ndarray)):
            ret = np.zeros((n,) + tuple(self._right_shape), dtype=self.dtype)
        elif isinstance(self.dtype, dict):  # rec array
            ret = np.zeros(n, dtype=self.dtype)
        else:
            raise ValueError("dtype not understood")
        return ret

    def append(self, data):
        # data = self.normalise_data(data)
        if len(self._buffer)==self.length:
            tmp = self._alloc(len(self._buffer) * 2)
            tmp[:self.length] = self._buffer
            self._buffer = tmp

        if isinstance(self.dtype, dict):
            for k, val in enumerate(data):
                self._buffer[self.length][k] = data[k]
        else:
            self._buffer[self.length] = data
            if self._ema is None:
                self._ema = self._buffer[self.length]
            else:
                f = self._ema_factor
                fc = 1 - f
                self._ema = self._ema * f + self._buffer[self.length] * fc

        self.length += 1

        self._min = np.minimum(data, self._min)
        self._max = np.maximum(data, self._max)
        self._sum = self._sum + np.asanyarray(data)

    def add_offset(self, off):
        self.data[:] += off
        if off.ndim>np.ndim(self._sum):
            off = off[0]
        self._min += off
        self._max += off
        self._sum += off * self.length

    def clear(self):
        self.length = 0
        self._min = +np.inf
        self._max = -np.inf
        self._sum = 0

    def mean(self):
        return np.asarray(self._sum, dtype=np.float32) / self.length

    def sum(self):
        return self._sum

    def max(self):
        return self._max

    def min(self):
        return self._min

    def __len__(self):
        return self.length

    @property
    def data(self):
        return self._buffer[:self.length]

    @property
    def ema(self):
        return self._ema

    def __getitem__(self, slc):
        return self._buffer[:self.length][slc]


class KDT(NearestNeighbors_):
    warning_shown = False

    @functools.wraps(NearestNeighbors_.__init__)
    def __init__(self, n_neighbors=5, radius=1.0, algorithm='auto',
                 leaf_size=30, metric='minkowski', p=2, metric_params=None,
                 n_jobs=1, **kwargs):
        if sklearn.__version__=="0.16.1":
            if not KDT.warning_shown:
                logger.warning("sklearn version does not support MP, try to upgrade it.")
                KDT.warning_shown = True

            if "n_jobs" in kwargs:
                kwargs.pop("n_jobs")
        else:
            kwargs['n_jobs'] = n_jobs

        super(KDT, self).__init__(n_neighbors=n_neighbors, radius=radius,
                                  algorithm=algorithm, leaf_size=leaf_size,
                                  metric=metric, p=p,
                                  metric_params=metric_params, **kwargs)

    __init__.__doc__ = NearestNeighbors_.__init__.__doc__


@numba.jit(nopython=True, looplift=True, cache=True)
def _merge(distances, indices, coordinates, pairwise_dist, sort_ix, new_points,
           k, query_points):
    q = len(query_points)
    dim = query_points.shape[1:]
    distances_new = np.zeros((q, k), dtype=np.float32)  # (q,k)
    indices_new = np.zeros((q, k), dtype=np.int64)  # (q,k)
    coordinates_new = np.zeros((q, k,) + dim, dtype=np.float32)  # (q,k,2/3)
    kdt_pointer = np.zeros(q, dtype=np.int64)  # (q) should be maximal k+1
    new_pointer = np.zeros(q, dtype=np.int64)  # (q) should be maximal m+1

    for p in range(q):  # over query points
        for c in range(k):  # over #NNs
            new_ix = sort_ix[p, new_pointer[p]]
            d_new = pairwise_dist[p, new_ix]
            d_kdt = distances[p, kdt_pointer[p]]
            if d_kdt>d_new:
                distances_new[p, c] = d_new
                indices_new[p, c] = -666
                coordinates_new[p, c] = new_points[new_ix]
                new_pointer[p] += 1
            else:
                distances_new[p, c] = d_kdt
                indices_new[p, c] = indices[p, kdt_pointer[p]]
                coordinates_new[p, c] = coordinates[p, kdt_pointer[p]]
                kdt_pointer[p] += 1

    return distances_new, indices_new, coordinates_new


class DynamicKDT(object):
    def __init__(self, points=None, k=1, n_jobs=-1, rebuild_thresh=100,
                 aniso_scale=None):
        self._kdt = None
        self._new_points = []
        self._static_points = []
        self._k = k
        self._jobs = n_jobs
        self._rebuild_thresh = rebuild_thresh
        self.aniso_scale = 1
        if aniso_scale is not None:
            if isinstance(aniso_scale, (list, tuple)):
                self.aniso_scale = np.atleast_2d(np.array(aniso_scale))
            elif isinstance(aniso_scale, np.ndarray):
                self.aniso_scale = np.atleast_2d(aniso_scale)
            else:
                raise ValueError("aniso_scale not understood")

        if points is not None:
            if len(points)<=k:
                raise ValueError("points must be longer than k")
            self._kdt = KDT(n_neighbors=k, n_jobs=n_jobs, algorithm='kd_tree',
                            leaf_size=20)

            self._kdt.fit(points * self.aniso_scale)
            self._static_points = points

    def append(self, point):
        point = np.asarray(point)
        if self._new_points==[]:
            self._new_points = AccumulationArray(right_shape=point.shape,
                                                 n_init=self._rebuild_thresh)

        if len(self._new_points)==self._rebuild_thresh:
            if self._static_points==[]:
                self._static_points = self._new_points.data.copy()
            else:
                self._static_points = np.concatenate(
                    [self._static_points, self._new_points.data], axis=0)
            self._new_points.clear()
            self._kdt = KDT(n_neighbors=self._k, n_jobs=self._jobs,
                            algorithm='kd_tree', leaf_size=20)
            self._kdt.fit(self._static_points * self.aniso_scale)

        self._new_points.append(point)

    def get_knn(self, query_points, k=None):
        if k is None:
            k = self._k

        if k>(len(self._new_points) + len(self._static_points)):
            raise ValueError("The requested number of neigbours is larger "
                             "than the number of stored points")
        if query_points.ndim==1:
            query_points = query_points[None]
        q = len(query_points)

        if len(self._static_points):
            # assert k==self._kdt.n_neighbors
            distances, indices = self._kdt.kneighbors(
                query_points * self.aniso_scale, n_neighbors=k)
            # Add inf for stopping
            distances = np.hstack([distances, np.ones((q, 1)) * np.inf])
        else:
            distances = np.ones((q, 1)) * np.inf
            indices = np.zeros((q, 1), dtype=np.int)

        if len(self._new_points):
            new_points = self._new_points.data
            pairwise_dist = cdist(query_points * self.aniso_scale,
                                  new_points * self.aniso_scale, p=2)
            # Add inf for stopping
            pairwise_dist = np.hstack(
                [pairwise_dist, np.ones((q, 1)) * np.inf])
        else:
            new_points = np.zeros((0, 1), dtype=query_points.dtype)
            pairwise_dist = np.ones((q, 1)) * np.inf  # (q,1)

        if k==1:
            indices = indices[:, 0]
            distances = distances[:, 0].astype(np.float32)
            if len(self._static_points):
                coordinates = self._static_points[indices]
            else:
                coordinates = new_points[indices]
            # Override found neighbours if a closer neighbour is in new_points
            replace_by_new = pairwise_dist.min(axis=1)<distances
            distances[replace_by_new] = pairwise_dist[replace_by_new]
            new_index = pairwise_dist.argmin(axis=1)
            if np.any(replace_by_new):
                coordinates[replace_by_new] = new_points[
                    new_index[replace_by_new]]
                indices[replace_by_new] = new_index[replace_by_new] + len(
                    self._static_points)  # -666 # This is just a dummy, atm indices is not used anyway
                distances[replace_by_new] = pairwise_dist[
                    replace_by_new, new_index[replace_by_new]]
        else:
            if len(self._static_points):
                coordinates = self._static_points[indices]
            else:
                coordinates = np.zeros_like(new_points)

            sort_ix = pairwise_dist.argsort(axis=1)  # (q,n) this is ascending
            distances, indices, coordinates = _merge(distances, indices,
                                                     coordinates,
                                                     pairwise_dist, sort_ix,
                                                     new_points, k,
                                                     query_points)

        if q==1:
            distances = distances[0]
            indices = indices[0]
            coordinates = coordinates[0]

        return distances, indices, coordinates

    def get_radius_nn(self, query_points, radius):
        raise NotImplementedError()


### Various Simple Utils ###

def import_variable_from_file(file_path, class_name):
    directory = os.path.dirname(file_path)
    sys.path.append(directory)
    mod_name = os.path.split(file_path)[1]
    if mod_name[-3:]=='.py':
        mod_name = mod_name[:-3]

    mod = importlib.import_module(mod_name)
    sys.path.pop(-1)
    cls = getattr(mod, class_name)
    return cls


def picklesave(data, file_name):
    """
    Writes one or many objects to pickle file

    data:
      single objects to save or iterable of objects to save.
      For iterable, all objects are written in this order to the file.
    file_name: string
      path/name of destination file
    """
    file_name = os.path.expanduser(file_name)
    with open(file_name, 'wb') as f:
        pkl.dump(data, f, protocol=2)


def pickleload(file_name):
    """
    Loads all object that are saved in the pickle file.
    Multiple objects are returned as list.
    """
    file_name = os.path.expanduser(file_name)
    ret = []
    try:
        with open(file_name, 'rb') as f:
            try:
                while True:
                    # Python 3 needs explicit encoding specification,
                    # which Python 2 lacks:
                    if sys.version_info.major>=3:
                        ret.append(pkl.load(f, encoding='latin1'))
                    else:
                        ret.append(pkl.load(f))
            except EOFError:
                pass

        if len(ret)==1:
            return ret[0]
        else:
            return ret

    except pkl.UnpicklingError:
        with gzip.open(file_name, 'rb') as f:
            try:
                while True:
                    # Python 3 needs explicit encoding specification,
                    # which Python 2 lacks:
                    if sys.version_info.major>=3:
                        ret.append(pkl.load(f, encoding='latin1'))
                    else:
                        ret.append(pkl.load(f))
            except EOFError:
                pass

        if len(ret)==1:
            return ret[0]
        else:
            return ret


def h5save(data, file_name, keys=None, compress=True):
    """
    Writes one or many arrays to h5 file

    data:
      single array to save or iterable of arrays to save.
      For iterable all arrays are written to the file.
    file_name: string
      path/name of destination file
    keys: string / list thereof
      For single arrays this is a single string which is used as a name
      for the data set.
      For multiple arrays each dataset is named by the corresponding key.
      If keys is ``None``, the dataset names created by enumeration: ``data%i``
    compress: Bool
      Whether to use lzf compression, defaults to ``True``. Most useful for
      label arrays.
    """
    file_name = os.path.expanduser(file_name)
    compr = 'lzf' if compress else None
    f = h5py.File(file_name, "w")
    if isinstance(data, list) or isinstance(data, tuple):
        if keys is not None:
            assert len(keys)==len(data)
        for i, d in enumerate(data):
            if keys is None:
                f.create_dataset(str(i), data=d, compression=compr)
            else:
                f.create_dataset(keys[i], data=d, compression=compr)
    else:
        if keys is None:
            f.create_dataset('data', data=data, compression=compr)
        else:
            f.create_dataset(keys, data=data, compression=compr)
    f.close()


def h5load(file_name, keys=None):
    """
    Loads data sets from h5 file

    file_name: string
      destination file
    keys: string / list thereof
      Load only data sets specified in keys and return as list in the order
      of ``keys``
      For a single key the data is returned directly - not as list
      If keys is ``None`` all datasets that are listed in the keys-attribute
      of the h5 file are loaded.
    """
    file_name = os.path.expanduser(file_name)
    ret = []
    try:
        f = h5py.File(file_name, "r")
    except IOError:
        raise IOError("Could not open h5-File %s" % (file_name))

    if keys is not None:
        try:
            if isinstance(keys, str):
                ret.append(f[keys].value)
            else:
                for k in keys:
                    ret.append(f[k].value)
        except KeyError:
            raise KeyError("Could not read h5-dataset named %s. Available "
                           "datasets: %s" % (keys, list(f.keys())))
    else:
        for k in f.keys():
            ret.append(f[k].value)

    f.close()

    if len(ret)==1:
        return ret[0]
    else:
        return ret


def pretty_string_ops(n):
    """
    Return a humanized string representation of a large number.
    """
    abbrevs = [(1000000000000, 'Tera Ops'),
               (1000000000, 'Giga Ops'),
               (1000000, 'Mega Ops'),
               (1000, 'kilo Ops')]
    for factor, suffix in abbrevs:
        if n>=factor:
            break
    return "%.1f %s" % (float(n) / factor, suffix)


def makeversiondir(path, dir_name=None, cd=False):
    path = os.path.expanduser(path)
    if dir_name:
        path = os.path.join(path, dir_name)
    while True:
        if os.path.exists(path):
            try:
                num = re.findall(r"-v(\d+)$", path)[0]
                num = int(num)
                i = 2 + np.int(np.log10(num) + 1)
                num = "-v" + str(int(num) + 1)
                path = path[:-i] + num
            except:
                path = path + "-v1"
        else:
            break

    os.makedirs(path, mode=0o755)
    if cd:
        os.chdir(path)
    return path


class Timer(object):
    def __init__(self, silent_all=False):
        self.last_t = time.time()
        self.total = 0
        self.checktimes = []
        self.checknames = []
        self.accumulator = {}
        self.silent_all = silent_all

    def check(self, name=None, silent=False):
        t = time.time()
        dt = t - self.last_t
        self.total += dt
        self.last_t = t
        name = name if name is not None else ""
        if not silent and not self.silent_all:
            print("%s\tdt=%.3g s,\tt=%.3g s" % (name, dt, self.total))

        self.checknames.append(name)
        self.checktimes.append(dt)
        if name is not None:
            accum = self.accumulator.get(name, 0)
            self.accumulator[name] = accum + dt

    def plot(self, accum=False):
        # I don't want this import every time utils is used
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        if accum:
            times = list(self.accumulator.values())
            names = list(self.accumulator.keys())
        else:
            times = self.checktimes
            names = self.checknames

        ind = np.arange(len(times))
        ax.bar(ind, times)
        ax.set_xticks(ind + 0.5)
        ax.set_xticklabels(names)
        plt.show()

    def summary(self, silent=False, print_func=None):
        s = "Total t: %.3gs\n" % self.total
        if len(self.accumulator):
            ix = np.argsort(self.accumulator.values())
            for i in ix:
                name, dt = self.accumulator.items()[i]
                s += "%s:\t\t%.3gs\t%.3g%%\n"\
                     %(name, dt, dt/self.total*100.0)
        else:
            ix = np.argsort(self.checktimes)
            for i in ix:
                name, dt = self.checknames[i], self.checktimes[i]
                s += "%s:\t\t%.3gs\t%.3g%%\n"\
                     % (name, dt. dt/self.total*100.0)
        if silent:
            return s
        else:
            if print_func:
                print_func(s)
            else:
                print(s)


def pretty_string_time(t):
    """Custom printing of elapsed time"""
    if t>4000:
        s = 't=%.1fh' % (t / 3600)
    elif t>300:
        s = 't=%.0fm' % (t / 60)
    else:
        s = 't=%.0fs' % (t)
    return s


def unique_rows(a):
    # removes duplicates from a
    a = np.ascontiguousarray(a)
    unique_a, index = np.unique(a.view([('', a.dtype)] * a.shape[1]),
                                return_index=True)
    ret = unique_a.view(a.dtype).reshape(
        (unique_a.shape[0], a.shape[1])), index
    print("Removed %i of %i (new %i)" % (
        len(a) - len(index), len(a), len(index)))
    return ret


def as_list(var):
    if var is None:
        return var
    elif isinstance(var, (list, tuple)):
        return list(var)
    else:
        return [var, ]



# -*- coding: utf-8 -*-
"""
Copyright (c) 2015 Marius Killinger, Sven Dorkenwald, Philipp Schubert
All rights reserved
"""
from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip

__all__ = ['Data', 'MNISTData', 'PianoData', 'PianoData_perc']

import hashlib
import logging
import os
import time
try:
    import urllib.request as urllib2
except ImportError:
    import urllib2

import numpy as np


logger = logging.getLogger('elektronn2log')

import sklearn

from ..utils import pickleload

class Data(object):
    """
    Load and prepare data, Base-Obj
    """
    def __init__(self, n_lab=None):
        self._pos           = 0
        # self.train_d = None
        # self.train_l = None
        # self.valid_d = None
        # self.valid_l = None
        # self.test_d = None
        # self.test_l = None

        if isinstance(self.train_d, np.ndarray):
            self._training_count = self.train_d.shape[0]
            if n_lab is None:
                self.n_lab = np.unique(self.train_l).size
            else:
                self.n_lab = n_lab
        elif isinstance(self.train_d, list):
            self._training_count = len(self.train_d)
            if n_lab is None:
                unique = [np.unique(l) for l in self.train_l]
                self.n_lab = np.unique(np.hstack(unique)).size
            else:
                self.n_lab = n_lab

        if self.example_shape is None:
            self.example_shape = self.train_d[0].shape
        self.n_ch = self.example_shape[0]

        self.rng = np.random.RandomState(np.uint32((time.time()*0.0001 - int(time.time()*0.0001))*4294967295))
        self.pid = os.getpid()
        logger.info(self.__repr__())
        self._perm = self.rng.permutation(self._training_count)


    def _reseed(self):
        """Reseeds the rng if the process ID has changed!"""
        current_pid = os.getpid()
        if current_pid!=self.pid:
            self.pid = current_pid
            self.rng.seed(np.uint32((time.time()*0.0001 - int(time.time()*0.0001))*4294967295+self.pid))
            logger.debug("Reseeding RNG in Process with PID: {}".format(self.pid))


    def __repr__(self):
        return "%i-class Data Set: #training examples: %i and #validing: %i" \
        %(self.n_lab, self._training_count, len(self.valid_d))


    def getbatch(self, batch_size, source='train'):
        if source=='train':
            if (self._pos+batch_size) < self._training_count:
                self._pos += batch_size
                slice = self._perm[self._pos-batch_size:self._pos]
            else: # get new permutation
                self._perm = self.rng.permutation(self._training_count)
                self._pos = 0
                slice = self._perm[:batch_size]

            if isinstance(self.train_d, np.ndarray):
                return (self.train_d[slice], self.train_l[slice])

            elif isinstance(self.train_d, list):
                data  = np.array([self.train_d[i] for i in slice])
                label = np.array([self.train_l[i] for i in slice])
                return (data, label)

        elif source=='valid':
            data  = self.valid_d[:batch_size]
            label = self.valid_l[:batch_size]
            return (data, label)

        elif source=='test':
            data  = self.test_d[:batch_size]
            label = self.test_l[:batch_size]
            return (data, label)

    def createCVSplit(self, data, label, n_folds=3, use_fold=2, shuffle=False, random_state=None):
        try:  # sklearn >=0.18 API
            # (see http://scikit-learn.org/dev/whats_new.html#model-selection-enhancements-and-api-changes)
            import sklearn.model_selection
            kfold = sklearn.model_selection.KFold(
                n_splits=n_folds, shuffle=shuffle, random_state=random_state
            )
            cv = kfold.split(data)
        except:  # sklearn <0.18 API # TODO: We can remove this after a while.
            import sklearn.cross_validation
            cv = sklearn.cross_validation.KFold(
                len(data), n_folds, shuffle=shuffle, random_state=random_state
            )
        for fold, (train_i, valid_i) in enumerate(cv):
            if fold==use_fold:
                self.valid_d = data[valid_i]
                self.valid_l = label[valid_i]
                self.train_d = data[train_i]
                self.train_l = label[train_i]

##########################################################################################

def _augmentMNIST(data, label, crop=2, factor=4):
    """
    Creates new data, by cropping/shifting data.
    Control blow-up by factor and maximum offset by crop
    """
    n = data.shape[-1]
    new_size    = (n-crop)
    new_data    = np.zeros((0,1,new_size,new_size), dtype=np.float32) # store new data in here
    new_label   = np.zeros((0,), dtype=np.int16)
    pos         = [(i%crop, int(i/crop)%crop) for i in range(crop**2)] # offests of different positions
    perm        = np.random.permutation(range(crop**2))

    for i in range(factor): # create <factor> new version of data
        ix =pos[perm[i]]
        new = (data[:, :, ix[0]:ix[0]+new_size, ix[1]:ix[1]+new_size])
        new_data = np.concatenate((new_data, new), axis=0)
        new_label= np.concatenate((new_label, label), axis=0)

    return new_data, new_label


class MNISTData(Data):
    def __init__(self, input_node, target_node,  path=None, convert2image=True,
                 warp_on=False, shift_augment=True, center=True):
        if path is None:
            (self.train_d, self.train_l), (self.valid_d, self.valid_l), (
            self.test_d, self.test_l) = self.download()

        else:
            path = os.path.expanduser(path)
            (self.train_d, self.train_l), (self.valid_d, self.valid_l), (self.test_d, self.test_l) = pickleload(path)

        self.warp_on = warp_on
        self.shif_augment = shift_augment
        self.return_flat  = not convert2image
        self.test_l  = self.test_l.astype(np.int16)
        self.train_l = self.train_l.astype(np.int16)
        self.valid_l = self.valid_l.astype(np.int16)
        self.example_shape = None

        if center:
            self.test_d  -= self.test_d.mean()
            self.train_d -= self.train_d.mean()
            self.valid_d -= self.valid_d.mean()

        self.convert_to_image()
        if self.shif_augment:
            self._stripborder(1)
            self.train_d, self.train_l = _augmentMNIST(self.train_d, self.train_l, crop=2, factor=4)

        self.train_l = self.train_l[:, None]
        self.test_l = self.test_l[:, None]
        self.valid_l = self.valid_l[:, None]
        super(MNISTData, self).__init__()
        if not convert2image:
            self.example_shape = self.train_d[0].size


        logger.info("MNIST data is converted/augmented to shape {}".format(self.example_shape))

    @staticmethod
    def download():
        if os.name == 'nt':
            dest = os.path.join(os.environ['APPDATA'], 'ELEKTRONN2')
        else:
            xdg_cache_home = os.environ.get('XDG_CACHE_HOME', os.path.expanduser('~/.cache'))
            dest = os.path.join(xdg_cache_home, 'ELEKTRONN2')

        if not os.path.exists(dest):
            os.makedirs(dest)

        dest = os.path.join(dest, 'mnist.pkl.gz')

        if os.path.exists(dest):
            print("Found existing MNIST data at {}".format(dest))
        else:
            download_url = "http://www.elektronn.org/downloads/mnist.pkl.gz"
            print("Downloading MNIST data from {}".format(download_url))
            # TODO: We could use a progress bar here.
            response = urllib2.urlopen(download_url)
            data = response.read()
            print("Saving data to {}".format(dest))
            with open(dest, "wb") as f:
                f.write(data)
            response.close()
            # Check if download is complete
            with open(dest, "rb") as f:
                mnist_sha256_abbr = '4b950cea3877c03ea8db5cb4'
                response_sha256 = hashlib.sha256(f.read()).hexdigest()
                if not response_sha256.startswith(mnist_sha256_abbr):
                    raise IOError('{} is corrupted. Please delete it and retry.'.format(dest))

        return pickleload(dest)

    def convert_to_image(self):
        """For MNIST / flattened 2d, single-Layer, square images"""

        valid_size  = self.valid_l.size
        test_size   = self.test_l.size
        data        = np.vstack((self.valid_d, self.test_d, self.train_d))
        size        = data[0].size
        n = int(np.sqrt(size))
        assert abs(n**2-size) < 1e-6 , '<convertToImage> data is not square'
        count = data.shape[0]
        data = data.reshape((count, 1, n, n))
        self.valid_d  = data[:valid_size]
        self.test_d = data[valid_size:valid_size+test_size]
        self.train_d = data[valid_size+test_size:]

    def getbatch(self,batch_size, source='train'):
        if source=='valid':
            ret = super(MNISTData, self).getbatch(batch_size, 'valid')
        if source=='test':
            ret =  super(MNISTData, self).getbatch(batch_size, 'test')
        else:
            d, l = super(MNISTData, self).getbatch(batch_size, source)
            if self.warp_on:
                d = self._warpaugment(d)
            ret = d, l

        if self.return_flat:
            ret = (ret[0].reshape((batch_size, -1)), ret[1])

        return ret

    def _stripborder(self, pix=1):
        s = self.train_d.shape[-1]
        self.valid_d = self.valid_d[:, :, pix:s-pix, pix:s-pix]
        self.test_d  = self.test_d [:, :, pix:s-pix, pix:s-pix]

    def _warpaugment(self, d, amount=1):
        rot_max     = 5  * amount
        shear_max   = 7   * amount
        scale_max   = 1.15 * amount
        stretch_max = 0.25 * amount

        shear = shear_max * 2 * (np.random.rand()-0.5)
        twist   = rot_max * 2 * (np.random.rand()-0.5)
        rot     = 0 # min(rot_max - abs(twist), rot_max  * (np.random.rand()))
        scale   = 1 + (scale_max-1) * np.random.rand(2)
        stretch =  stretch_max * 2 * (np.random.rand(4)-0.5)

        ps = (d.shape[0],)+d.shape[2:]
        raise ValueError("Warping is suspended, reimplement using warp.py")
        #w =  warping.warp3dFast(d, ps, rot, shear, (scale[0], scale[1], 1), stretch, twist)
        return w


class PianoData(Data):
  def __init__(self, input_node, target_node,
               path='/home/mkilling/devel/data/PianoRoll/Nottingham_enc.pkl', n_tap=20, n_lab=58):
    path = os.path.expanduser(path)
    (self.train_d, self.valid_d, self.test_d)  = pickleload(path)
    super(PianoData, self).__init__(n_lab=n_lab)
    self.example_shape = self.train_d[0].shape[-1]
    self.n_taps = n_tap
    self.n_lab = n_lab


  def getbatch(self, batch_size, source='train'):
    if source=='train':
      if (self._pos+batch_size) < self._training_count:
        self._pos += batch_size
        slice = self._perm[self._pos-batch_size:self._pos]
      else: # get new permutation
        self._perm = self.rng.permutation(self._training_count)
        self._pos = 0
        slice = self._perm[:batch_size]

      data  = [self.train_d[i] for i in slice]

    elif source=='valid':
      data  = self.valid_d[:batch_size]

    elif source=='test':
      data  = self.test_d[:batch_size]

    lengths = np.array(map(len, data))
    start_t = np.round(np.random.rand(batch_size)*(lengths-self.n_taps-1)).astype(np.int)
    x = np.array([d[t:t+self.n_taps].astype(np.float32) for d,t in zip(data, start_t)])
    y = np.array([d[t+self.n_taps]                      for d,t in zip(data, start_t)])
    return x, y


class PianoData_perc(PianoData):
  def __init__(self, input_node, target_node,
               path='/home/mkilling/devel/data/PianoRoll/Nottingham_enc.pkl', n_tap=20, n_lab=58):
    super(PianoData_perc, self).__init__(input_node, target_node, path='/home/mkilling/devel/data/PianoRoll/Nottingham_enc.pkl', n_tap=20, n_lab=58)


  def getbatch(self, batch_size, source='train'):
    x, y = super(PianoData_perc, self).getbatch(batch_size, source)
    x = x.swapaxes(0, 1)
    return x.reshape((x.shape[0], -1)), y
